from typing import Type

from vedro.core import ConfigType, Dispatcher, Plugin, PluginConfig
from vedro.events import ArgParsedEvent, ConfigLoadedEvent

from ._scenario_assert_rewriter_loader import ScenarioAssertRewriterLoader

__all__ = ("AssertRewriter", "AssertRewriterPlugin",)


class AssertRewriterPlugin(Plugin):
    def __init__(self, config: Type["AssertRewriter"]):
        super().__init__(config)

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ConfigLoadedEvent, self.on_config_loaded) \
                  .listen(ArgParsedEvent, self.on_arg_parsed)

    def on_config_loaded(self, event: ConfigLoadedEvent) -> None:
        self._global_config: ConfigType = event.config

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        try:
            from dessert import AssertionRewritingHook  # type: ignore
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Package 'dessert' is not found, install it via 'pip install dessert'")

        self._global_config.Registry.ScenarioLoader.register(
            lambda: ScenarioAssertRewriterLoader(AssertionRewritingHook()), self)


class AssertRewriter(PluginConfig):
    plugin = AssertRewriterPlugin
    description = "Rewrites assert statements to provide better error messages"



from typing import List

from .._virtual_scenario import VirtualScenario
from ._scenario_orderer import ScenarioOrderer

__all__ = ("PlainScenarioOrderer",)


# deprecated
class PlainScenarioOrderer(ScenarioOrderer):
    async def sort(self, scenarios: List[VirtualScenario]) -> List[VirtualScenario]:
        return scenarios



import os
import sys
from pathlib import Path
from typing import List, Union

from vedro.core import Config

from ._config_generator import ConfigGenerator
from ._config_markup import ConfigMarkup
from ._config_parser import ConfigParser

__all__ = ("ConfigUpdater",)


class ConfigUpdater:
    def __init__(self, config_path: Path) -> None:
        self._config_path = config_path
        self._default_config_path = Path("vedro.cfg.py")

    async def update(self, plugin_package: str, plugin_name: str, *, enabled: bool) -> None:
        assert sys.version_info >= (3, 8), "Python 3.8+ required"

        config_path = self._get_config_path()
        config_source = await self._read_config(config_path)
        config_parser = ConfigParser()
        config_markup = await config_parser.parse(config_source)
        config_generator = ConfigGenerator(config_markup.get_indent())

        if not config_markup.get_config_section():
            imports = self._gen_imports_if_needed(["vedro", plugin_package],
                                                  config_markup, config_generator)
            generated = config_generator.gen_config_section(
                config_generator.gen_plugins_section(
                    config_generator.gen_plugin_section(plugin_package, plugin_name, enabled)
                )
            )
            last_import = self._get_last_import(config_markup) or 0
            start_lineno = last_import + 1
            replace = False

        elif not config_markup.get_plugin_list_section():
            imports = self._gen_imports_if_needed(["vedro", plugin_package],
                                                  config_markup, config_generator)
            generated = config_generator.gen_plugins_section(
                config_generator.gen_plugin_section(plugin_package, plugin_name, enabled)
            )
            config_section = config_markup.get_config_section()
            assert config_section is not None
            start_lineno = config_section["end"] + 1
            replace = False

        elif not config_markup.get_plugin_section(plugin_name):
            imports = self._gen_imports_if_needed([plugin_package],
                                                  config_markup, config_generator)
            generated = config_generator.gen_plugin_section(plugin_package, plugin_name, enabled)
            plugin_list_section = config_markup.get_plugin_list_section()
            assert plugin_list_section is not None
            start_lineno = plugin_list_section["end"] + 1
            replace = False

        elif not config_markup.get_enabled_attr(plugin_name):
            imports = []
            generated = config_generator.gen_enabled_attr(enabled)
            plugin_section = config_markup.get_plugin_section(plugin_name)
            assert plugin_section is not None
            start_lineno = plugin_section["end"] + 1
            replace = False

        else:
            imports = []
            generated = config_generator.gen_enabled_attr(enabled)
            enabled_attr = config_markup.get_enabled_attr(plugin_name)
            assert enabled_attr is not None
            start_lineno = enabled_attr["start"]
            replace = True

        config_source = self.apply(config_source, generated, start_lineno, replace)
        config_source = self.apply(config_source, imports, 1)
        await self._save_config(config_path, config_source)

    def _gen_imports_if_needed(self, imports: List[str],
                               markup: ConfigMarkup,
                               generator: ConfigGenerator) -> List[str]:
        generated = []
        for import_name in imports:
            import_info = markup.get_import(import_name)
            if (import_info is None) or (import_info["alias"] is not None):
                generated += generator.gen_import(import_name)

        last_import = self._get_last_import(markup)
        if len(generated) > 0 and (last_import is None):
            return generated + [os.linesep]
        return generated

    def _get_last_import(self, markup: ConfigMarkup) -> Union[int, None]:
        config_section = markup.get_config_section()
        last_import = -1
        for import_info in markup.get_import_list():
            if config_section and (import_info["start"] > config_section["start"]):
                continue
            if import_info["start"] > last_import:
                last_import = import_info["start"]

        return last_import if (last_import > 0) else None

    def apply(self, config_source: str, generated: List[str], lineno: int, replace: bool = False,
              *, linesep: str = os.linesep) -> str:
        config_lines = config_source.split(linesep)

        if lineno > len(config_lines):
            for _ in range(lineno - len(config_lines)):
                config_lines.append("")

        new_config_lines = []
        for num, line in enumerate(config_lines, start=1):
            if num == lineno:
                new_config_lines.extend(generated)
                if not replace:
                    new_config_lines.append(line)
            else:
                new_config_lines.append(line)

        return linesep.join(new_config_lines)

    def _get_config_path(self) -> Path:
        path = self._config_path.absolute()
        if path == Config.path:
            return self._default_config_path
        return path

    async def _read_config(self, config_path: Path) -> str:
        try:
            with open(config_path, "r") as f:
                return f.read()
        except FileNotFoundError:
            return ""

    async def _save_config(self, config_path: Path, config_source: str) -> None:
        with open(config_path, "w") as f:
            f.write(config_source)



import sys
from time import time
from typing import List, Tuple, Type

from ..._scenario import Scenario
from ...events import (
    ExceptionRaisedEvent,
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    ScenarioReportedEvent,
    ScenarioRunEvent,
    ScenarioSkippedEvent,
    StepFailedEvent,
    StepPassedEvent,
    StepRunEvent,
)
from .._dispatcher import Dispatcher
from .._exc_info import ExcInfo
from .._report import Report
from .._scenario_result import ScenarioResult
from .._step_result import StepResult
from .._virtual_scenario import VirtualScenario
from .._virtual_step import VirtualStep
from ..scenario_scheduler import ScenarioScheduler
from ._interrupted import Interrupted, RunInterrupted, ScenarioInterrupted, StepInterrupted
from ._scenario_runner import ScenarioRunner

__all__ = ("MonotonicScenarioRunner",)


class MonotonicScenarioRunner(ScenarioRunner):
    def __init__(self, dispatcher: Dispatcher, *,
                 interrupt_exceptions: Tuple[Type[BaseException], ...] = ()) -> None:
        self._dispatcher = dispatcher
        assert isinstance(interrupt_exceptions, tuple)
        self._interrupt_exceptions = interrupt_exceptions + (Interrupted,)

    def _is_interruption(self, exc_info: ExcInfo,
                         exceptions: Tuple[Type[BaseException], ...]) -> bool:
        for exception in exceptions:
            if isinstance(exc_info.value, exception):
                return True
        return False

    async def run_step(self, step: VirtualStep, ref: Scenario) -> StepResult:
        step_result = StepResult(step)

        await self._dispatcher.fire(StepRunEvent(step_result))
        step_result.set_started_at(time())
        try:
            if step.is_coro():
                await step(ref)
            else:
                step(ref)
        except:  # noqa: E722
            step_result.set_ended_at(time()).mark_failed()

            exc_info = ExcInfo(*sys.exc_info())
            await self._dispatcher.fire(ExceptionRaisedEvent(exc_info))
            step_result.set_exc_info(exc_info)

            await self._dispatcher.fire(StepFailedEvent(step_result))

            if self._is_interruption(exc_info, self._interrupt_exceptions):
                raise StepInterrupted(exc_info, step_result)
        else:
            step_result.set_ended_at(time()).mark_passed()
            await self._dispatcher.fire(StepPassedEvent(step_result))

        return step_result

    async def run_scenario(self, scenario: VirtualScenario) -> ScenarioResult:
        scenario_result = ScenarioResult(scenario)

        if scenario.is_skipped():
            scenario_result.mark_skipped()
            await self._dispatcher.fire(ScenarioSkippedEvent(scenario_result))
            return scenario_result

        await self._dispatcher.fire(ScenarioRunEvent(scenario_result))
        scenario_result.set_started_at(time())

        ref = scenario()
        scenario_result.set_scope(ref.__dict__)

        for step in scenario.steps:
            try:
                step_result = await self.run_step(step, ref)
            except self._interrupt_exceptions as e:
                if isinstance(e, StepInterrupted):
                    scenario_result.add_step_result(e.step_result)
                    exc_info = e.exc_info
                else:
                    exc_info = ExcInfo(*sys.exc_info())
                scenario_result.set_ended_at(time()).mark_failed()
                await self._dispatcher.fire(ScenarioFailedEvent(scenario_result))
                raise ScenarioInterrupted(exc_info, scenario_result)
            else:
                scenario_result.add_step_result(step_result)

            if step_result.is_failed():
                scenario_result.set_ended_at(time()).mark_failed()
                await self._dispatcher.fire(ScenarioFailedEvent(scenario_result))
                break

        if not scenario_result.is_failed():
            scenario_result.set_ended_at(time()).mark_passed()
            await self._dispatcher.fire(ScenarioPassedEvent(scenario_result))

        return scenario_result

    async def _report_scenario_results(self, scenario_results: List[ScenarioResult],
                                       report: Report, scheduler: ScenarioScheduler) -> None:
        aggregated_result = scheduler.aggregate_results(scenario_results)
        report.add_result(aggregated_result)
        await self._dispatcher.fire(ScenarioReportedEvent(aggregated_result))

    async def _run_scenarios(self, scheduler: ScenarioScheduler, report: Report) -> None:
        scenario_results: List[ScenarioResult] = []

        async for scenario in scheduler:
            prev_scenario = scenario_results[-1].scenario if len(scenario_results) > 0 else None
            if prev_scenario and prev_scenario.unique_id != scenario.unique_id:
                await self._report_scenario_results(scenario_results, report, scheduler)
                scenario_results = []

            try:
                scenario_result = await self.run_scenario(scenario)
            except self._interrupt_exceptions as e:
                if isinstance(e, ScenarioInterrupted):
                    scenario_results.append(e.scenario_result)
                    exc_info = e.exc_info
                else:
                    exc_info = ExcInfo(*sys.exc_info())
                if len(scenario_results) > 0:
                    await self._report_scenario_results(scenario_results, report, scheduler)
                raise RunInterrupted(exc_info)
            else:
                scenario_results.append(scenario_result)

        if len(scenario_results) > 0:
            await self._report_scenario_results(scenario_results, report, scheduler)

    async def run(self, scheduler: ScenarioScheduler) -> Report:
        report = Report()
        try:
            await self._run_scenarios(scheduler, report)
        except self._interrupt_exceptions as e:
            exc_info = e.exc_info if isinstance(e, RunInterrupted) else ExcInfo(*sys.exc_info())
            report.set_interrupted(exc_info)
        return report



import asyncio
from asyncio import create_subprocess_shell
from asyncio.streams import StreamReader
from typing import Any, Callable, Union

__all__ = ("install_plugin",)

CallbackType = Callable[[str], Any]


async def install_plugin(package: str, pip_args: str, *,
                         on_stdout: CallbackType, on_stderr: CallbackType) -> Union[int, None]:
    cmd = f"pip install {package} {pip_args}"

    proc = await create_subprocess_shell(cmd.strip(),
                                         stdout=asyncio.subprocess.PIPE,
                                         stderr=asyncio.subprocess.PIPE)
    await asyncio.gather(
        _read_stream(proc.stdout, on_stdout),
        _read_stream(proc.stderr, on_stderr)
    )

    await proc.wait()

    return proc.returncode


async def _read_stream(stream: Union[StreamReader, None], callback: CallbackType) -> None:
    if not stream:
        return

    while True:
        line = await stream.readline()
        if not line:
            break
        callback(line.decode())



from pathlib import Path
from typing import Any, List, Tuple

from .._virtual_scenario import VirtualScenario
from ._scenario_orderer import ScenarioOrderer

__all__ = ("StableScenarioOrderer",)


class StableScenarioOrderer(ScenarioOrderer):
    def _cmp(self, scn: VirtualScenario) -> Tuple[Any, ...]:
        path = Path(scn.path)
        return (len(path.parts),) + tuple((len(x), x) for x in path.parts)

    async def sort(self, scenarios: List[VirtualScenario]) -> List[VirtualScenario]:
        return list(sorted(scenarios, key=self._cmp))



import re
from contextlib import redirect_stderr, redirect_stdout
from importlib import import_module
from inspect import isclass
from io import StringIO
from pathlib import Path
from types import ModuleType
from typing import List, Tuple, Union

from vedro.core import PluginConfig

from ._config_updater import ConfigUpdater

__all__ = ("PluginManager",)


class PluginManager:
    def __init__(self, config_path: Path) -> None:
        self._config_path = config_path
        self._config_updater = ConfigUpdater(config_path)

    async def enable(self, plugin_name: str) -> List[Tuple[str, str]]:
        return await self.toogle(plugin_name, enabled=True)

    async def disable(self, plugin_name: str) -> List[Tuple[str, str]]:
        return await self.toogle(plugin_name, enabled=False)

    async def toogle(self, plugin_name: str, *, enabled: bool) -> List[Tuple[str, str]]:
        ext_package = plugin_name.replace("-", "_")
        core_package = f"vedro.plugins.{self._to_snake_case(plugin_name)}"
        plugins = self._get_plugins(ext_package) or self._get_plugins(core_package)
        for plugin_package, plugin_name in plugins:
            await self._config_updater.update(plugin_package, plugin_name, enabled=enabled)
        return plugins

    def _to_snake_case(self, name: str) -> str:
        name = re.sub("(.)([A-Z][a-z]+)", r"1_\2", name)
        return re.sub("([a-z0-9])([A-Z])", r"1_\2", name).lower()

    def _get_plugins(self, plugin_package: str) -> List[Tuple[str, str]]:
        plugins: List[Tuple[str, str]] = []

        plugin_package = plugin_package.replace("-", "_")
        module = self._import_module(plugin_package)
        if module is None:
            return plugins

        for key, val in module.__dict__.items():
            if not key.startswith("_") and isclass(val) and issubclass(val, PluginConfig):
                plugins.append((plugin_package, key))
        return plugins

    def _import_module(self, module_name: str) -> Union[ModuleType, None]:
        with StringIO() as buf:
            with redirect_stdout(buf), redirect_stderr(buf):
                try:
                    return import_module(module_name)
                except ImportError:
                    return None



import json
import platform
import urllib.request
from typing import Dict, List, cast

import vedro

__all__ = ("fetch_top_plugins",)


async def fetch_top_plugins(limit: int, timeout: float, *,
                            api_url: str = "https://api.vedro.io/v1") -> List[Dict[str, str]]:
    url = api_url + f"/plugins/top?limit={limit}"
    headers = {"User-Agent": _get_user_agent()}
    plugins = _send_request(url, headers=headers, timeout=timeout)
    return plugins


def _send_request(url: str, *, headers: Dict[str, str], timeout: float) -> List[Dict[str, str]]:
    request = urllib.request.Request(url, headers=headers)
    with urllib.request.urlopen(request, timeout=timeout) as response:
        result = json.loads(response.read())
    return cast(List[Dict[str, str]], result)


def _get_user_agent() -> str:
    python_version = platform.python_version()
    platform_info = platform.platform(terse=True)
    return f"Vedro/{vedro.__version__} (Python/{python_version}; {platform_info})"



from typing import Any, Callable, Dict, Type

from vedro.core import Dispatcher, PluginConfig, ScenarioResult
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    ScenarioRunEvent,
    ScenarioSkippedEvent,
)

from .._director_init_event import DirectorInitEvent
from .._reporter import Reporter
from ..rich import RichPrinter

__all__ = ("PyCharmReporter", "PyCharmReporterPlugin",)


class PyCharmReporterPlugin(Reporter):
    def __init__(self, config: Type["PyCharmReporter"], *,
                 printer_factory: Callable[[], RichPrinter] = RichPrinter) -> None:
        super().__init__(config)
        self._printer = printer_factory()
        self._show_internal_calls = config.show_internal_calls
        self._show_skipped = config.show_skipped
        self._no_output = config.no_output

    def subscribe(self, dispatcher: Dispatcher) -> None:
        super().subscribe(dispatcher)
        dispatcher.listen(DirectorInitEvent, lambda e: e.director.register("pycharm", self))

    def on_chosen(self) -> None:
        assert isinstance(self._dispatcher, Dispatcher)
        self._dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                        .listen(ArgParsedEvent, self.on_arg_parsed) \
                        .listen(ScenarioRunEvent, self.on_scenario_run) \
                        .listen(ScenarioPassedEvent, self.on_scenario_passed) \
                        .listen(ScenarioFailedEvent, self.on_scenario_failed) \
                        .listen(ScenarioSkippedEvent, self.on_scenario_skipped) \
                        .listen(CleanupEvent, self.on_cleanup)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        group = event.arg_parser.add_argument_group("PyCharm Reporter")

        group.add_argument("--pycharm-no-output",
                           action="store_true",
                           default=self._no_output,
                           help="Don't produce report output")
        group.add_argument("--pycharm-show-skipped",
                           action="store_true",
                           default=self._show_skipped,
                           help="Show skipped scenarios (deprecated, default: True)")
        group.add_argument("--pycharm-show-internal-calls",
                           action="store_true",
                           default=self._show_internal_calls,
                           help="Show internal calls in the traceback output")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._show_internal_calls = event.args.pycharm_show_internal_calls
        self._show_skipped = event.args.pycharm_show_skipped
        self._no_output = event.args.pycharm_no_output

    def on_scenario_run(self, event: ScenarioRunEvent) -> None:
        scenario_result = event.scenario_result
        self._write_message("testStarted", {
            "name": scenario_result.scenario.subject,
            "locationHint": scenario_result.scenario.path,
        })

    def _print_scenario(self, scenario_result: ScenarioResult) -> None:
        if self._no_output:
            return

        self._printer.print_scenario_subject(scenario_result.scenario.subject,
                                             scenario_result.status, prefix=" ")

        if not scenario_result.is_failed():
            return

        for step_result in scenario_result.step_results:
            self._printer.print_step_name(step_result.step_name, step_result.status,
                                          prefix=" " * 3)
            if step_result.exc_info:
                self._printer.print_exception(step_result.exc_info,
                                              show_internal_calls=self._show_internal_calls)

        if scenario_result.scope:
            self._printer.print_scope(scenario_result.scope)

    def on_scenario_passed(self, event: ScenarioPassedEvent) -> None:
        scenario_result = event.scenario_result
        self._print_scenario(scenario_result)
        self._write_message("testFinished", {
            "name": scenario_result.scenario.subject,
            "duration": int(scenario_result.elapsed * 1000),
        })

    def on_scenario_failed(self, event: ScenarioFailedEvent) -> None:
        scenario_result = event.scenario_result
        self._print_scenario(scenario_result)
        self._write_message("testFailed", {
            "name": scenario_result.scenario.subject,
            "message": "",
            "details": "",
        })
        self._write_message("testFinished", {
            "name": scenario_result.scenario.subject,
            "duration": int(scenario_result.elapsed * 1000),
        })

    def on_scenario_skipped(self, event: ScenarioSkippedEvent) -> None:
        if not self._show_skipped:
            return
        scenario_result = event.scenario_result
        self._print_scenario(scenario_result)
        self._write_message("testIgnored", {
            "name": scenario_result.scenario.subject,
        })

    def on_cleanup(self, event: CleanupEvent) -> None:
        if self._no_output:
            return

        if event.report.interrupted:
            self._printer.print_interrupted(event.report.interrupted)

    def _escape_value(self, value: str) -> str:
        symbols = {"'": "|'", "\n": "|n", "\r": "|r", "|": "||", "[": "|[", "]": "|]"}
        return value.translate({ord(k): v for k, v in symbols.items()})

    def _write_message(self, name: str, attributes: Dict[str, Any]) -> None:
        message = f"##teamcity[{name}"
        for key, val in attributes.items():
            escaped = self._escape_value(str(val))
            message += f" {key}='{escaped}'"
        message += "]"
        self._printer.console.out(message)


class PyCharmReporter(PluginConfig):
    plugin = PyCharmReporterPlugin
    description = "Outputs scenario results in a PyCharm-friendly format"

    # Show internal calls in the traceback output
    show_internal_calls: bool = False

    # Show skipped scenarios
    show_skipped: bool = True

    # Don't produce report output
    no_output: bool = False



from typing import Any, Dict, List, Union

from vedro.core._artifacts import Artifact
from vedro.core._step_result import StepResult
from vedro.core._virtual_scenario import VirtualScenario

from ._scenario_status import ScenarioStatus

__all__ = ("ScenarioResult",)

ScopeType = Dict[str, Any]


class ScenarioResult:
    def __init__(self, scenario: VirtualScenario) -> None:
        self._scenario = scenario
        self._status: ScenarioStatus = ScenarioStatus.PENDING
        self._started_at: Union[float, None] = None
        self._ended_at: Union[float, None] = None
        self._step_results: List[StepResult] = []
        self._scope: Union[ScopeType, None] = None
        self._artifacts: List[Artifact] = []

    @property
    def scenario(self) -> VirtualScenario:
        return self._scenario

    @property
    def status(self) -> ScenarioStatus:
        return self._status

    def mark_passed(self) -> "ScenarioResult":
        self._status = ScenarioStatus.PASSED
        return self

    def is_passed(self) -> bool:
        return self._status == ScenarioStatus.PASSED

    def mark_failed(self) -> "ScenarioResult":
        self._status = ScenarioStatus.FAILED
        return self

    def is_failed(self) -> bool:
        return self._status == ScenarioStatus.FAILED

    def mark_skipped(self) -> "ScenarioResult":
        self._status = ScenarioStatus.SKIPPED
        return self

    def is_skipped(self) -> bool:
        return self._status == ScenarioStatus.SKIPPED

    @property
    def started_at(self) -> Union[float, None]:
        return self._started_at

    def set_started_at(self, started_at: float) -> "ScenarioResult":
        self._started_at = started_at
        return self

    @property
    def ended_at(self) -> Union[float, None]:
        return self._ended_at

    def set_ended_at(self, ended_at: float) -> "ScenarioResult":
        self._ended_at = ended_at
        return self

    @property
    def elapsed(self) -> float:
        if (self._started_at is None) or (self._ended_at is None):
            return 0.0
        return self._ended_at - self._started_at

    def add_step_result(self, step_result: StepResult) -> None:
        self._step_results.append(step_result)

    @property
    def step_results(self) -> List[StepResult]:
        return self._step_results[:]

    def set_scope(self, scope: ScopeType) -> None:
        self._scope = scope

    @property
    def scope(self) -> ScopeType:
        if self._scope is None:
            return {}
        return self._scope

    def attach(self, artifact: Artifact) -> None:
        assert isinstance(artifact, Artifact)
        self._artifacts.append(artifact)

    @property
    def artifacts(self) -> List[Artifact]:
        return self._artifacts[:]

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self._scenario!r} {self._status.value}>"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and (self.__dict__ == other.__dict__)



from asyncio import iscoroutinefunction
from collections import deque
from typing import Any, Callable, Deque, Dict, Tuple, Type, Union

from vedro.core import Dispatcher, Plugin, PluginConfig
from vedro.events import ScenarioFailedEvent, ScenarioPassedEvent, ScenarioRunEvent

__all__ = ("Deferrer", "DeferrerPlugin", "defer", "Deferrable",)


Deferrable = Tuple[Callable[..., Any], Tuple[Any, ...], Dict[str, Any]]

_queue: Deque[Deferrable] = deque()


def defer(fn: Callable[..., Any], *args: Any, **kwargs: Any) -> None:
    _queue.append((fn, args, kwargs))


class DeferrerPlugin(Plugin):
    def __init__(self, config: Type["Deferrer"], *, queue: Deque[Deferrable] = _queue) -> None:
        super().__init__(config)
        self._queue = queue

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ScenarioRunEvent, self.on_scenario_run) \
                  .listen(ScenarioPassedEvent, self.on_scenario_end) \
                  .listen(ScenarioFailedEvent, self.on_scenario_end)

    def on_scenario_run(self, event: ScenarioRunEvent) -> None:
        self._queue.clear()

    async def on_scenario_end(self,
                              event: Union[ScenarioPassedEvent, ScenarioFailedEvent]) -> None:
        while len(self._queue) > 0:
            fn, args, kwargs = self._queue.pop()
            if iscoroutinefunction(fn):
                await fn(*args, **kwargs)
            else:
                fn(*args, **kwargs)


class Deferrer(PluginConfig):
    plugin = DeferrerPlugin
    description = "Executes deferred functions at the end of each scenario"



import sys
from typing import Dict, List, Union

__all__ = ("ConfigMarkup", "ConfigSectionType", "PluginListSectionType",
           "PluginSectionType", "EnabledAttrType", "ImportType", "MarkupElement",)


if sys.version_info >= (3, 8):
    from typing import TypedDict

    class MarkupElement(TypedDict):
        offset: int
        start: int
        end: int

    class ImportType(MarkupElement):
        alias: Union[str, None]

    class EnabledAttrType(MarkupElement):
        pass

    class PluginSectionType(MarkupElement):
        enabled: Union[EnabledAttrType, None]

    class PluginListSectionType(MarkupElement):
        children: Dict[str, PluginSectionType]

    class ConfigSectionType(MarkupElement):
        plugins: Union[PluginListSectionType, None]

else:  # pragma: no cover
    from typing import Any
    MarkupElement = Dict[str, Any]
    ImportType = Dict[str, Any]
    EnabledAttrType = Dict[str, Any]
    PluginSectionType = Dict[str, Any]
    PluginListSectionType = Dict[str, Any]
    ConfigSectionType = Dict[str, Any]


class ConfigMarkup:
    def __init__(self, config: Union[ConfigSectionType, None],
                 imports: Dict[str, ImportType],
                 indent: str) -> None:
        self._config_section = config
        self._imports = imports
        self._indent = indent

    def get_indent(self) -> str:
        return self._indent

    def get_import_list(self) -> List[ImportType]:
        return list(self._imports.values())

    def get_import(self, package_name: str) -> Union[ImportType, None]:
        return self._imports.get(package_name, None)

    def get_config_section(self) -> Union[ConfigSectionType, None]:
        return self._config_section

    def get_plugin_list_section(self) -> Union[PluginListSectionType, None]:
        config_section = self.get_config_section()
        return config_section["plugins"] if config_section else None

    def get_plugin_section(self, plugin_name: str) -> Union[PluginSectionType, None]:
        plugin_list_section = self.get_plugin_list_section()
        return plugin_list_section["children"].get(plugin_name) if plugin_list_section else None

    def get_enabled_attr(self, plugin_name: str) -> Union[EnabledAttrType, None]:
        plugin_section = self.get_plugin_section(plugin_name)
        return plugin_section["enabled"] if plugin_section else None



from ._interrupted import Interrupted, RunInterrupted, ScenarioInterrupted, StepInterrupted
from ._monotonic_scenario_runner import MonotonicScenarioRunner
from ._scenario_runner import ScenarioRunner

__all__ = ("ScenarioRunner", "MonotonicScenarioRunner",
           "Interrupted", "StepInterrupted", "ScenarioInterrupted", "RunInterrupted",)



from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Iterator, List

from .._scenario_result import AggregatedResult, ScenarioResult
from .._virtual_scenario import VirtualScenario

__all__ = ("ScenarioScheduler",)


class ScenarioScheduler(ABC):
    def __init__(self, scenarios: List[VirtualScenario]) -> None:
        self._discovered = OrderedDict((scn.unique_id, scn) for scn in scenarios)

    @property
    def discovered(self) -> Iterator[VirtualScenario]:
        for scenario_id in self._discovered:
            yield self._discovered[scenario_id]

    @property
    @abstractmethod
    def scheduled(self) -> Iterator[VirtualScenario]:
        pass

    @abstractmethod
    def schedule(self, scenario: VirtualScenario) -> None:
        pass

    @abstractmethod
    def ignore(self, scenario: VirtualScenario) -> None:
        pass

    @abstractmethod
    def aggregate_results(self, scenario_results: List[ScenarioResult]) -> AggregatedResult:
        pass

    def __aiter__(self) -> "ScenarioScheduler":
        return self

    @abstractmethod
    async def __anext__(self) -> VirtualScenario:
        pass

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"



from inspect import isclass
from typing import Callable, Type, TypeVar, overload

from vedro._scenario import Scenario

__all__ = ("skip",)


T = TypeVar("T", bound=Type[Scenario])


@overload
def skip(scenario_or_reason: T) -> T:  # pragma: no cover
    pass


@overload
def skip(scenario_or_reason: str) -> Callable[[T], T]:  # pragma: no cover
    pass


@overload
def skip(scenario_or_reason: None = None) -> Callable[[T], T]:  # pragma: no cover
    pass


def skip(scenario_or_reason=None):  # type: ignore
    def wrapped(scenario: T) -> T:
        if not issubclass(scenario, Scenario):
            raise TypeError("Decorator @skip can be used only with 'vedro.Scenario' subclasses")

        setattr(scenario, "__vedro__skipped__", True)
        if isinstance(scenario_or_reason, str):
            setattr(scenario, "__vedro__skip_reason__", scenario_or_reason)

        return scenario

    if (scenario_or_reason is None) or isinstance(scenario_or_reason, str):
        return wrapped
    elif isclass(scenario_or_reason):
        return wrapped(scenario_or_reason)
    else:
        raise TypeError('Usage: @skip or @skip("reason")')



from pathlib import Path
from typing import List

from ._file_filter import FileFilter

__all__ = ("AnyFilter",)


class AnyFilter(FileFilter):
    def __init__(self, filters: List[FileFilter]) -> None:
        self._filters = filters

    def filter(self, path: Path) -> bool:
        for filter_ in self._filters:
            if filter_.filter(path):
                return True
        return False



import sys
from dataclasses import dataclass
from typing import Callable, List, Type

from rich.console import Console
from rich.status import Status
from rich.table import Table

from vedro import Config

from .._cmd_arg_parser import CommandArgumentParser
from .._command import Command
from ._fetch_plugins import fetch_top_plugins
from ._get_plugin_info import get_plugin_info
from ._install_plugin import install_plugin
from .plugin_manager import PluginManager

__all__ = ("PluginCommand", "PluginInfo",)


def make_console() -> Console:
    return Console(highlight=False, force_terminal=True, markup=False, soft_wrap=True)


@dataclass
class PluginInfo:
    name: str
    enabled: bool
    package: str = "Unknown"
    version: str = "0.0.0"
    summary: str = "No data"
    is_default: bool = False


class PluginCommand(Command):
    def __init__(self, config: Type[Config], arg_parser: CommandArgumentParser, *,
                 console_factory: Callable[[], Console] = make_console) -> None:
        super().__init__(config, arg_parser)
        self._console = console_factory()
        self._plugin_manager = PluginManager(config.path)

    async def run(self) -> None:
        subparsers = self._arg_parser.add_subparsers(dest="subparser")

        subparsers.add_parser("list", help="Show installed plugins")
        subparsers.add_parser("top", help="Show popular plugins")

        install_subparser = subparsers.add_parser("install", help="Install plugin")
        install_subparser.add_argument("packages", nargs="+", help="Package name(s)")
        install_subparser.add_argument("--pip-args", default="", help="Additional pip arguments")

        enable_subparser = subparsers.add_parser("enable", help="Enable plugin")
        enable_subparser.add_argument("plugin", help="Plugin name")

        disable_subparser = subparsers.add_parser("disable", help="Disable plugin")
        disable_subparser.add_argument("plugin", help="Plugin name")

        args = self._arg_parser.parse_args()
        if args.subparser == "top":
            await self._show_top_plugins()
        elif args.subparser == "list":
            await self._show_installed_plugins()
        elif args.subparser == "install":
            if sys.version_info < (3, 8):
                self._console.print("Python 3.8+ required for this command", style="red")
                return
            await self._install_packages(args.packages, args.pip_args)
        elif args.subparser == "enable":
            await self._enable_plugin(args.plugin)
        elif args.subparser == "disable":
            await self._disable_plugin(args.plugin)
        else:
            self._arg_parser.print_help()
            self._arg_parser.exit()

    async def _show_top_plugins(self) -> None:
        with self._console.status("Fetching plugins..."):
            try:
                plugins = await fetch_top_plugins(limit=10, timeout=10.0)
            except Exception as e:
                self._console.print(f"Failed to fetch popular plugins ({e})", style="red")
                return

        table = Table(expand=True, border_style="grey50")
        table.add_column("Package", overflow="fold", style="blue")
        table.add_column("Description", overflow="fold")
        table.add_column("URL", overflow="fold")
        table.add_column("Popularity", justify="right")

        for plugin in plugins:
            table.add_row(plugin["name"], plugin["description"],
                          plugin["url"], str(plugin["popularity"]))

        self._console.print(table)

    async def _install_packages(self, packages: List[str], pip_args: str) -> None:
        for package in packages:
            await self._install_package(package, pip_args)

    async def _install_package(self, package: str, pip_args: str) -> None:
        message = f"Installing '{package}'..."
        stdout, stderr = [], []

        def on_stdout(line: str, st: Status) -> None:
            stdout.append(line)
            last_line = "".join(stdout[-1:])
            st.update(f"{message}\n[grey50]{last_line}[/]")

        def on_stderr(line: str, st: Status) -> None:
            stderr.append(line)
            last_line = "".join(stderr[-1:])
            st.update(f"{message}\n[yellow]{last_line}[/]")

        with self._console.status(message) as status:
            return_code = await install_plugin(
                package=package,
                pip_args=pip_args,
                on_stdout=lambda line: on_stdout(line, status),
                on_stderr=lambda line: on_stderr(line, status),
            )

        if return_code == 0:
            enabled = await self._plugin_manager.enable(package)
            self._console.print(f"✔ Successfully installed '{package}'", style="green")
            if len(enabled) == 0:
                self._console.print(f"✗ Failed to enable '{package}'", style="red")
        else:
            self._console.print(f"✗ Failed to install '{package}'", style="red")
            self._console.print("".join(stdout), style="grey50", end="")
            self._console.print("".join(stderr), style="yellow", end="")

    async def _enable_plugin(self, plugin_name: str) -> None:
        enabled = await self._plugin_manager.enable(plugin_name)
        if len(enabled) == 0:
            self._console.print(f"✗ Plugin '{plugin_name}' not found", style="red")
        else:
            self._console.print(f"✔ Plugin '{plugin_name}' enabled", style="green")

    async def _disable_plugin(self, plugin_name: str) -> None:
        disabled = await self._plugin_manager.disable(plugin_name)
        if len(disabled) == 0:
            self._console.print(f"✗ Plugin '{plugin_name}' not found", style="red")
        else:
            self._console.print(f"✔ Plugin '{plugin_name}' disabled", style="green")

    async def _show_installed_plugins(self) -> None:
        table = Table(expand=True, border_style="grey50")
        for column in ("Name", "Package", "Description", "Version", "Enabled"):
            table.add_column(column)

        for plugin_config in self._config.Plugins.values():
            plugin_info = get_plugin_info(plugin_config)

            color = "blue" if plugin_config.enabled else "grey70"
            table.add_row(plugin_info.name, plugin_info.package, plugin_info.summary,
                          plugin_info.version, str(plugin_info.enabled), style=color)

        self._console.print(table)



import json
import platform
import urllib.request
from threading import Thread
from typing import Any, Dict, Tuple, Type, Union

import vedro
from vedro.core import Dispatcher, Plugin, PluginConfig
from vedro.events import CleanupEvent, StartupEvent

__all__ = ("SystemUpgrade", "SystemUpgradePlugin",)


class SystemUpgradePlugin(Plugin):
    def __init__(self, config: Type["SystemUpgrade"]) -> None:
        super().__init__(config)
        self._api_url = config.api_url
        self._timeout = config.timeout
        self._thread: Union[Thread, None] = None
        self._latest_version: Union[str, None] = None
        self._cur_version = vedro.__version__

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(StartupEvent, self.on_startup) \
                  .listen(CleanupEvent, self.on_cleanup)

    def _send_request(self, url: str, *,
                      headers: Dict[str, str], timeout: float) -> Union[Any, None]:
        request = urllib.request.Request(url, headers=headers)
        try:
            with urllib.request.urlopen(request, timeout=timeout) as response:
                return json.loads(response.read())
        except BaseException:
            return None

    def _get_user_agent(self) -> str:
        python_version = platform.python_version()
        platform_info = platform.platform(terse=True)
        return f"Vedro/{self._cur_version} (Python/{python_version}; {platform_info})"

    def _get_latest_version(self) -> None:
        url = f"{self._api_url}/v1/latest-version"
        headers = {"User-Agent": self._get_user_agent()}
        response = self._send_request(url, headers=headers, timeout=self._timeout)
        if isinstance(response, dict) and ("version" in response):
            self._latest_version = response["version"]

    def on_startup(self, event: StartupEvent) -> None:
        self._thread = Thread(target=self._get_latest_version, daemon=True)
        self._thread.start()

    def _is_up_to_date(self, cur_version: str, new_version: str) -> bool:
        try:
            from pkg_resources import parse_version  # type: ignore
        except ImportError:
            def parse_version(version: str) -> Tuple[str, ...]:
                return tuple(x.zfill(8) for x in version.split("."))
        return bool(parse_version(cur_version) >= parse_version(new_version))

    def on_cleanup(self, event: CleanupEvent) -> None:
        if self._thread:
            self._thread.join(0.0)
            self._thread = None

        if self._latest_version is None:
            return

        if not self._is_up_to_date(self._cur_version, self._latest_version):
            message = f"(!) Vedro update available: {self._cur_version} → {self._latest_version}"
            event.report.add_summary(message)


class SystemUpgrade(PluginConfig):
    plugin = SystemUpgradePlugin
    description = "Checks for Vedro updates and notifies when a newer version is available"

    # URL to the Vedro API
    api_url: str = "https://api.vedro.io"

    # Timeout for the request to the API
    timeout: float = 1.0



from typing import Type

from ._config_loader import Section
from ._dispatcher import Dispatcher, Subscriber

__all__ = ("Plugin", "PluginConfig",)


class Plugin(Subscriber):
    def __init__(self, config: Type["PluginConfig"]) -> None:
        assert issubclass(config, PluginConfig)
        self._config = config

    @property
    def config(self) -> Type["PluginConfig"]:
        return self._config

    def subscribe(self, dispatcher: Dispatcher) -> None:
        pass

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._config.__name__})"


class PluginConfig(Section):
    plugin: Type[Plugin] = Plugin
    description: str = ""
    enabled: bool = True



from typing import List

from vedro.core import VirtualScenario
from vedro.core.scenario_orderer import StableScenarioOrderer

__all__ = ("ReversedOrderer",)


class ReversedOrderer(StableScenarioOrderer):
    async def sort(self, scenarios: List[VirtualScenario]) -> List[VirtualScenario]:
        return list(reversed(await super().sort(scenarios)))



from abc import ABC, abstractmethod
from typing import Any, Callable, Generic, Type, TypeVar, Union

from ._plugin import Plugin

__all__ = ("Container", "Factory", "Singleton", "FactoryType", "ConflictError")

F = TypeVar("F")
FactoryType = Union[Type[F], Callable[..., F]]

T = TypeVar("T")


class ConflictError(Exception):
    pass


class Container(Generic[T], ABC):
    def __init__(self, resolver: FactoryType[T]) -> None:
        self._resolver = resolver
        self._initial = resolver
        self._registrant: Union[Plugin, None] = None

    def _make_conflict_error(self, registrant: Plugin) -> ConflictError:
        type_ = self.__orig_class__.__args__[0]  # type: ignore
        return ConflictError(f"{registrant} is trying to register {type_.__name__}, "
                             f"but it is already registered by {self._registrant!r}")

    @abstractmethod
    def register(self, resolver: FactoryType[T], registrant: Plugin) -> None:
        pass

    @abstractmethod
    def resolve(self, *args: Any, **kwargs: Any) -> T:
        pass

    def __call__(self, *args: Any, **kwargs: Any) -> T:
        return self.resolve(*args, **kwargs)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._resolver!r})"


class Factory(Container[T]):
    def register(self, resolver: FactoryType[T], registrant: Plugin) -> None:
        assert isinstance(registrant, Plugin)

        if self._registrant is not None:
            raise self._make_conflict_error(registrant)

        self._resolver = resolver
        self._registrant = registrant

    def resolve(self, *args: Any, **kwargs: Any) -> T:
        return self._resolver(*args, **kwargs)


class Singleton(Container[T]):
    def __init__(self, resolver: FactoryType[T]) -> None:
        super().__init__(resolver)
        self._singleton: Union[None, T] = None

    def register(self, resolver: FactoryType[T], registrant: Plugin) -> None:
        assert isinstance(registrant, Plugin)

        if self._registrant is not None:
            raise self._make_conflict_error(registrant)

        self._resolver = resolver
        self._registrant = registrant

    def resolve(self, *args: Any, **kwargs: Any) -> T:
        if self._singleton is None:
            self._singleton = self._resolver(*args, **kwargs)
        return self._singleton



from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Tuple, TypeVar, Union, cast

__all__ = ("params",)


class Parameterized:
    def __init__(self, args: Any, kwargs: Any, decorators: Tuple[Any, ...]) -> None:
        self._args = args
        self._kwargs = kwargs
        self._decorators = decorators

    def __call__(self, fn: Callable[..., None]) -> Callable[..., None]:
        if not hasattr(fn, "__vedro__params__"):
            setattr(fn, "__vedro__params__", [])
        getattr(fn, "__vedro__params__").append((self._args, self._kwargs, self._decorators))
        return fn


T = TypeVar("T")
ItemType = Union[Callable[[T], T], Tuple[Callable[[T], T], ...]]


class Params:
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._args = args
        self._kwargs = kwargs

    def __call__(self, fn: Callable[..., None]) -> Callable[..., None]:
        return Parameterized(self._args, self._kwargs, ())(fn)

    def __class_getitem__(cls, item: ItemType[T]) -> Callable[..., Parameterized]:
        decorators = item if isinstance(item, tuple) else (item,)

        @wraps(cls)
        def wrapped(*args: Any, **kwargs: Any) -> Parameterized:
            return Parameterized(args, kwargs, decorators)

        return wrapped


if not TYPE_CHECKING:
    params = Params
else:  # pragma: no cover
    # https://github.com/python/mypy/issues/11501
    # So we have to use TYPE_CHECKING here
    _T = TypeVar("_T")

    class TypedParams:
        def __call__(self, *args: Any, **kwargs: Any) -> Callable[[_T], _T]:
            return cast(Callable[[_T], _T], ...)

        def __getitem__(self, item: ItemType[T]) -> Callable[..., Callable[[_T], _T]]:
            return cast(Callable[..., Callable[[_T], _T]], ...)
    params = TypedParams()



import random
from typing import List

from vedro.core import ScenarioOrderer, VirtualScenario

__all__ = ("RandomOrderer",)


class RandomOrderer(ScenarioOrderer):
    async def sort(self, scenarios: List[VirtualScenario]) -> List[VirtualScenario]:
        copied = scenarios[:]
        random.shuffle(copied)
        return copied



from ._assert_rewriter import AssertRewriter, AssertRewriterPlugin
from ._scenario_assert_rewriter_loader import ScenarioAssertRewriterLoader

__all__ = ("AssertRewriter", "AssertRewriterPlugin", "ScenarioAssertRewriterLoader",)



from collections import OrderedDict
from typing import Iterator, List, Tuple

from .._scenario_result import AggregatedResult, ScenarioResult
from .._virtual_scenario import VirtualScenario
from ._scenario_scheduler import ScenarioScheduler

__all__ = ("MonotonicScenarioScheduler",)


class MonotonicScenarioScheduler(ScenarioScheduler):
    def __init__(self, scenarios: List[VirtualScenario]) -> None:
        super().__init__(scenarios)
        self._scheduled = OrderedDict((k, (v, 0)) for k, v in reversed(self._discovered.items()))
        self._queue: OrderedDict[str, Tuple[VirtualScenario, int]] = OrderedDict()

    @property
    def scheduled(self) -> Iterator[VirtualScenario]:
        for scenario_id in reversed(self._scheduled):
            scenario, repeats = self._scheduled[scenario_id]
            for _ in range(repeats + 1):
                yield scenario

    def ignore(self, scenario: VirtualScenario) -> None:
        self._scheduled.pop(scenario.unique_id, None)
        self._queue.pop(scenario.unique_id, None)

    def __aiter__(self) -> "ScenarioScheduler":
        self._queue = self._scheduled.copy()
        return super().__aiter__()

    async def __anext__(self) -> VirtualScenario:
        while len(self._queue) > 0:
            _, (scenario, repeats) = self._queue.popitem()
            if repeats > 0:
                self._queue[scenario.unique_id] = (scenario, repeats - 1)
            return scenario
        raise StopAsyncIteration()

    def schedule(self, scenario: VirtualScenario) -> None:
        if scenario.unique_id in self._scheduled:
            scn, repeats = self._scheduled[scenario.unique_id]
            scheduled = (scn, repeats + 1)
        else:
            scheduled = (scenario, 0)
        self._scheduled[scenario.unique_id] = scheduled

        if scenario.unique_id in self._queue:
            scn, repeats = self._queue[scenario.unique_id]
            queued = (scn, repeats + 1)
        else:
            queued = (scenario, 0)
        self._queue[scenario.unique_id] = queued

    def aggregate_results(self, scenario_results: List[ScenarioResult]) -> AggregatedResult:
        assert len(scenario_results) > 0

        result = scenario_results[0]
        for scenario_result in scenario_results:
            if scenario_result.is_failed():
                result = scenario_result
                break

        return AggregatedResult.from_existing(result, scenario_results)



from typing import Type, Union

from vedro.core import ConfigType, Dispatcher, Plugin, PluginConfig, ScenarioScheduler
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ConfigLoadedEvent,
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    StartupEvent,
)

from ._scheduler import RerunnerScenarioScheduler

__all__ = ("Rerunner", "RerunnerPlugin",)


class RerunnerPlugin(Plugin):
    def __init__(self, config: Type["Rerunner"]) -> None:
        super().__init__(config)
        self._reruns: int = 0
        self._global_config: Union[ConfigType, None] = None
        self._scheduler: Union[ScenarioScheduler, None] = None
        self._rerun_scenario_id: Union[str, None] = None
        self._reran: int = 0
        self._times: int = 0

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ConfigLoadedEvent, self.on_config_loaded) \
                  .listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(StartupEvent, self.on_startup) \
                  .listen(ScenarioPassedEvent, self.on_scenario_end) \
                  .listen(ScenarioFailedEvent, self.on_scenario_end) \
                  .listen(CleanupEvent, self.on_cleanup)

    def on_config_loaded(self, event: ConfigLoadedEvent) -> None:
        self._global_config = event.config

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        group = event.arg_parser.add_argument_group("Rerunner")
        group.add_argument("--reruns", type=int, default=0,
                           help="Number of times to rerun failed scenarios (default: 0)")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._reruns = event.args.reruns
        assert self._reruns >= 0

        if self._reruns == 0:
            return

        assert self._global_config is not None
        self._global_config.Registry.ScenarioScheduler.register(RerunnerScenarioScheduler, self)

    def on_startup(self, event: StartupEvent) -> None:
        self._scheduler = event.scheduler

    def on_scenario_end(self, event:  Union[ScenarioPassedEvent, ScenarioFailedEvent]) -> None:
        if self._reruns == 0:
            return

        assert isinstance(self._scheduler, RerunnerScenarioScheduler)

        if self._rerun_scenario_id == event.scenario_result.scenario.unique_id:
            return

        self._rerun_scenario_id = event.scenario_result.scenario.unique_id
        if event.scenario_result.is_failed():
            self._reran += 1
            for _ in range(self._reruns):
                self._scheduler.schedule(event.scenario_result.scenario)
                self._times += 1

    def on_cleanup(self, event: CleanupEvent) -> None:
        if self._reruns != 0:
            ss = "" if self._reran == 1 else "s"
            ts = "" if self._times == 1 else "s"
            event.report.add_summary(f"rerun {self._reran} scenario{ss}, {self._times} time{ts}")


class Rerunner(PluginConfig):
    plugin = RerunnerPlugin
    description = "Reruns failed scenarios a specified number of times"



from enum import Enum
from typing import Any, List, Union

from ._artifacts import Artifact
from ._exc_info import ExcInfo
from ._virtual_step import VirtualStep

__all__ = ("StepResult", "StepStatus",)


class StepStatus(Enum):
    PENDING = "PENDING"
    PASSED = "PASSED"
    FAILED = "FAILED"


class StepResult:
    def __init__(self, step: VirtualStep) -> None:
        self._step = step
        self._status: StepStatus = StepStatus.PENDING
        self._started_at: Union[float, None] = None
        self._ended_at: Union[float, None] = None
        self._exc_info: Union[ExcInfo, None] = None
        self._artifacts: List[Artifact] = []

    @property
    def step_name(self) -> str:
        return self._step.name

    @property
    def status(self) -> StepStatus:
        return self._status

    def is_passed(self) -> bool:
        return self._status == StepStatus.PASSED

    def is_failed(self) -> bool:
        return self._status == StepStatus.FAILED

    def mark_failed(self) -> "StepResult":
        self._status = StepStatus.FAILED
        return self

    def mark_passed(self) -> "StepResult":
        self._status = StepStatus.PASSED
        return self

    @property
    def started_at(self) -> Union[float, None]:
        return self._started_at

    def set_started_at(self, started_at: float) -> "StepResult":
        self._started_at = started_at
        return self

    @property
    def ended_at(self) -> Union[float, None]:
        return self._ended_at

    def set_ended_at(self, ended_at: float) -> "StepResult":
        self._ended_at = ended_at
        return self

    @property
    def elapsed(self) -> float:
        if (self._started_at is None) or (self._ended_at is None):
            return 0.0
        return self._ended_at - self._started_at

    @property
    def exc_info(self) -> Union[ExcInfo, None]:
        return self._exc_info

    def set_exc_info(self, exc_info: ExcInfo) -> "StepResult":
        self._exc_info = exc_info
        return self

    def attach(self, artifact: Artifact) -> None:
        assert isinstance(artifact, Artifact)
        self._artifacts.append(artifact)

    @property
    def artifacts(self) -> List[Artifact]:
        return self._artifacts[:]

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self._step!r} {self._status.value}>"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and (self.__dict__ == other.__dict__)



from pathlib import Path

from ._file_filter import FileFilter

__all__ = ("DunderFilter",)


class DunderFilter(FileFilter):
    def filter(self, path: Path) -> bool:
        p = path
        while p.suffix != "":
            p = p.with_suffix("")
        stem = p.stem
        return stem.startswith("__") and stem.endswith("__")



import asyncio
import os
import sys
from typing import Any

from ._config import Config
from ._context import context
from ._interface import Interface
from ._main import main
from ._params import params
from ._scenario import Scenario
from ._version import version
from .plugins.deferrer import defer
from .plugins.skipper import only, skip, skip_if

__version__ = version
__all__ = ("Scenario", "Interface", "run", "only", "skip", "skip_if", "params",
           "context", "defer", "Config",)


def run(*, plugins: Any = None) -> None:
    if plugins is not None:
        raise DeprecationWarning("Argument 'plugins' is deprecated, "
                                 "declare plugins in config (vedro.cfg.py)")

    cwd = os.getcwd()
    if cwd not in sys.path:
        sys.path.insert(0, cwd)

    asyncio.run(main())



import inspect
from typing import Type

from ..._scenario import Scenario
from .._virtual_scenario import VirtualScenario
from .._virtual_step import VirtualStep

__all__ = ("create_vscenario",)


def create_vscenario(scenario: Type[Scenario]) -> VirtualScenario:
    steps = []
    for step in scenario.__dict__:
        if step.startswith("_"):
            continue
        method = getattr(scenario, step)
        if not inspect.isfunction(method):
            continue
        steps.append(VirtualStep(method))
    return VirtualScenario(scenario, steps)



import warnings
from pathlib import Path
from typing import List, Type

from ..._scenario import Scenario
from .._virtual_scenario import VirtualScenario
from .._virtual_step import VirtualStep
from ._create_vscenario import create_vscenario
from ._scenario_discoverer import ScenarioDiscoverer

__all__ = ("MultiScenarioDiscoverer",)


class MultiScenarioDiscoverer(ScenarioDiscoverer):
    def _discover_steps(self, scenario: Type[Scenario]) -> List[VirtualStep]:
        # child classes may use this method to discover steps
        warnings.warn("Deprecated: use create_vscenario instead", DeprecationWarning)
        return create_vscenario(scenario).steps

    async def discover(self, root: Path) -> List[VirtualScenario]:
        scenarios = []
        async for path in self._finder.find(root):
            loaded = await self._loader.load(path)
            for scn in loaded:
                scenarios.append(create_vscenario(scn))
        return await self._orderer.sort(scenarios)



from typing import TYPE_CHECKING, Any

from vedro.core import Event

DirectorPlugin = Any
if TYPE_CHECKING:  # pragma: no cover
    from ._director import DirectorPlugin

__all__ = ("DirectorInitEvent",)


class DirectorInitEvent(Event):
    def __init__(self, director: DirectorPlugin) -> None:
        self._director = director

    @property
    def director(self) -> DirectorPlugin:
        return self._director

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._director!r})"



from typing import Callable, Type

from rich.console import Console
from rich.style import Style

import vedro
from vedro import Config

from .._cmd_arg_parser import CommandArgumentParser
from .._command import Command

__all__ = ("VersionCommand",)


def make_console() -> Console:
    return Console(highlight=False, force_terminal=True, markup=False, soft_wrap=True)


class VersionCommand(Command):
    def __init__(self, config: Type[Config], arg_parser: CommandArgumentParser, *,
                 console_factory: Callable[[], Console] = make_console) -> None:
        super().__init__(config, arg_parser)
        self._console = console_factory()

    async def run(self) -> None:
        self._arg_parser.parse_args()
        self._console.print(f"Vedro {vedro.__version__}", style=Style(color="blue"))



from abc import ABC, abstractmethod
from asyncio import iscoroutinefunction
from heapq import heappop, heappush
from time import monotonic_ns
from typing import Any, Callable, Dict, List, Type

from ._event import Event

__all__ = ("Dispatcher", "Subscriber", "EventHandler",)


HandlerType = Callable[..., Any]


class EventHandler:
    def __init__(self, priority: int, registered_at: int, handler: HandlerType) -> None:
        self._priority = priority
        self._registered_at = registered_at
        self._handler = handler

    async def __call__(self, event: Event) -> None:
        if iscoroutinefunction(self._handler):
            await self._handler(event)
        else:
            self._handler(event)

    def __lt__(self, other: "EventHandler") -> bool:
        assert isinstance(other, EventHandler)
        if self._priority == other._priority:
            return self._registered_at < other._registered_at
        return self._priority < other._priority


# backward compatibility
RegisteredItemType = EventHandler


class Subscriber(ABC):
    @abstractmethod
    def subscribe(self, dispatcher: "Dispatcher") -> None:
        raise NotImplementedError()


class Dispatcher:
    def __init__(self) -> None:
        self._events: Dict[str, List[EventHandler]] = {}

    def register(self, subscriber: "Subscriber") -> None:
        subscriber.subscribe(self)

    def listen(self, event: Type[Event], handler: HandlerType, priority: int = 0) -> "Dispatcher":
        assert issubclass(event, Event), "Event must be a subclass of 'vedro.events.Event'"
        if event.__name__ not in self._events:
            self._events[event.__name__] = []
        heappush(self._events[event.__name__], EventHandler(priority, monotonic_ns(), handler))
        return self

    async def fire(self, event: Event) -> None:
        if event.__class__.__name__ not in self._events:
            return
        registered = self._events[event.__class__.__name__]
        registered_copy: List[EventHandler] = []
        while len(registered) > 0:
            handler = heappop(registered)
            await handler(event)
            heappush(registered_copy, handler)
        self._events[event.__class__.__name__] = registered_copy



from .._exc_info import ExcInfo
from .._scenario_result import ScenarioResult
from .._step_result import StepResult

__all__ = ("Interrupted", "StepInterrupted", "ScenarioInterrupted", "RunInterrupted")


class Interrupted(BaseException):
    pass


class StepInterrupted(Interrupted):
    def __init__(self, exc_info: ExcInfo, step_result: StepResult) -> None:
        self._exc_info = exc_info
        self._step_result = step_result

    @property
    def exc_info(self) -> ExcInfo:
        return self._exc_info

    @property
    def step_result(self) -> StepResult:
        return self._step_result

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._exc_info!r}, {self._step_result!r})"


class ScenarioInterrupted(Interrupted):
    def __init__(self, exc_info: ExcInfo, scenario_result: ScenarioResult) -> None:
        self._exc_info = exc_info
        self._scenario_result = scenario_result

    @property
    def exc_info(self) -> ExcInfo:
        return self._exc_info

    @property
    def scenario_result(self) -> ScenarioResult:
        return self._scenario_result

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._exc_info!r}, {self._scenario_result!r})"


class RunInterrupted(Interrupted):
    def __init__(self, exc_info: ExcInfo) -> None:
        self._exc_info = exc_info

    @property
    def exc_info(self) -> ExcInfo:
        return self._exc_info

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._exc_info!r})"



from abc import ABC
from pathlib import Path
from typing import Any

__all__ = ("Artifact", "MemoryArtifact", "FileArtifact",)


class Artifact(ABC):
    pass


class MemoryArtifact(Artifact):
    def __init__(self, name: str,  mime_type: str, data: bytes) -> None:
        assert isinstance(data, bytes)
        self._name = name
        self._data = data
        self._mime_type = mime_type

    @property
    def name(self) -> str:
        return self._name

    @property
    def mime_type(self) -> str:
        return self._mime_type

    @property
    def data(self) -> bytes:
        return self._data

    def __repr__(self) -> str:
        size = len(self._data)
        return f"{self.__class__.__name__}<{self._name!r}, {self._mime_type!r}, size={size}>"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and (self.__dict__ == other.__dict__)


class FileArtifact(Artifact):
    def __init__(self, name: str, mime_type: str, path: Path) -> None:
        assert isinstance(path, Path)
        self._name = name
        self._path = path
        self._mime_type = mime_type

    @property
    def name(self) -> str:
        return self._name

    @property
    def mime_type(self) -> str:
        return self._mime_type

    @property
    def path(self) -> Path:
        return self._path

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<{self._name!r}, {self._mime_type!r}, {self._path!r}>"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and (self.__dict__ == other.__dict__)



import sys
from time import time
from typing import List, Tuple, Type

from vedro.core import (
    Dispatcher,
    ExcInfo,
    Report,
    ScenarioResult,
    ScenarioScheduler,
    StepResult,
    VirtualScenario,
    VirtualStep,
)
from vedro.core.scenario_runner import Interrupted, ScenarioRunner
from vedro.events import (
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    ScenarioReportedEvent,
    ScenarioRunEvent,
    ScenarioSkippedEvent,
    StepPassedEvent,
    StepRunEvent,
)

__all__ = ("DryRunner",)


class DryRunner(ScenarioRunner):
    def __init__(self, dispatcher: Dispatcher, *,
                 interrupt_exceptions: Tuple[Type[BaseException], ...] = ()) -> None:
        self._dispatcher = dispatcher
        assert isinstance(interrupt_exceptions, tuple)
        self._interrupt_exceptions = interrupt_exceptions + (Interrupted,)

    async def run_step(self, step: VirtualStep) -> StepResult:
        step_result = StepResult(step)

        await self._dispatcher.fire(StepRunEvent(step_result))
        step_result.set_started_at(time()).set_ended_at(time())
        step_result.mark_passed()
        await self._dispatcher.fire(StepPassedEvent(step_result))

        return step_result

    async def run_scenario(self, scenario: VirtualScenario) -> ScenarioResult:
        scenario_result = ScenarioResult(scenario)

        if scenario.is_skipped():
            scenario_result.mark_skipped()
            await self._dispatcher.fire(ScenarioSkippedEvent(scenario_result))
            return scenario_result

        await self._dispatcher.fire(ScenarioRunEvent(scenario_result))
        scenario_result.set_started_at(time())

        ref = scenario()
        scenario_result.set_scope(ref.__dict__)

        for step in scenario.steps:
            step_result = await self.run_step(step)
            scenario_result.add_step_result(step_result)

            if step_result.is_failed():
                scenario_result.set_ended_at(time())
                scenario_result.mark_failed()
                await self._dispatcher.fire(ScenarioFailedEvent(scenario_result))
                break

        if not scenario_result.is_failed():
            scenario_result.set_ended_at(time())
            scenario_result.mark_passed()
            await self._dispatcher.fire(ScenarioPassedEvent(scenario_result))

        return scenario_result

    async def _report_scenario_results(self, scenario_results: List[ScenarioResult],
                                       report: Report, scheduler: ScenarioScheduler) -> None:
        aggregated_result = scheduler.aggregate_results(scenario_results)
        report.add_result(aggregated_result)
        await self._dispatcher.fire(ScenarioReportedEvent(aggregated_result))

    async def _run_scenarios(self, scheduler: ScenarioScheduler, report: Report) -> None:
        scenario_results: List[ScenarioResult] = []

        async for scenario in scheduler:
            prev_scenario = scenario_results[-1].scenario if len(scenario_results) > 0 else None
            if prev_scenario and prev_scenario.unique_id != scenario.unique_id:
                await self._report_scenario_results(scenario_results, report, scheduler)
                scenario_results = []

            scenario_result = await self.run_scenario(scenario)
            scenario_results.append(scenario_result)

        if len(scenario_results) > 0:
            await self._report_scenario_results(scenario_results, report, scheduler)

    async def run(self, scheduler: ScenarioScheduler) -> Report:
        report = Report()
        try:
            await self._run_scenarios(scheduler, report)
        except self._interrupt_exceptions:
            exc_info = ExcInfo(*sys.exc_info())
            report.set_interrupted(exc_info)
        return report



import sys
from argparse import ArgumentParser, HelpFormatter
from functools import partial
from pathlib import Path
from typing import Type, cast

from ._config import Config
from .commands import CommandArgumentParser
from .commands.plugin_command import PluginCommand
from .commands.run_command import RunCommand
from .commands.version_command import VersionCommand
from .core import ConfigFileLoader


async def main() -> None:
    formatter = partial(HelpFormatter, max_help_position=30)
    config_path = Path("vedro.cfg.py")

    arg_parser = ArgumentParser("vedro", formatter_class=formatter,
                                add_help=False, allow_abbrev=False,
                                description="documentation: vedro.io/docs")

    config_loader = ConfigFileLoader(Config)
    config = cast(Type[Config], await config_loader.load(config_path))

    commands = ["run", "version", "plugin"]
    arg_parser.add_argument("command", nargs="?", help=f"Command to run {{{', '.join(commands)}}}")
    args, unknown_args = arg_parser.parse_known_args()

    # backward compatibility
    # vedro <args> -> vedro run <args>
    help_args = {"-h", "--help"}
    if (args.command not in commands) and (not help_args.intersection(set(unknown_args))):
        default_command = "run"
        sys.argv.insert(1, default_command)
        args.command = default_command

    arg_parser_factory = partial(CommandArgumentParser, formatter_class=formatter,
                                 allow_abbrev=False, add_help=True)

    if args.command == "run":
        parser = arg_parser_factory("vedro run")
        await RunCommand(config, parser).run()

    elif args.command == "version":
        parser = arg_parser_factory("vedro version")
        await VersionCommand(config, parser).run()

    elif args.command == "plugin":
        parser = arg_parser_factory("vedro plugin")
        await PluginCommand(config, parser).run()

    else:
        arg_parser.print_help()
        arg_parser.exit()



import argparse
import sys
from argparse import ArgumentParser, Namespace
from typing import List, Optional, Sequence, Tuple

__all__ = ("CommandArgumentParser",)


class CommandArgumentParser(ArgumentParser):
    def parse_known_args(self, args: Optional[Sequence[str]] = None,
                         namespace: Optional[Namespace] = None) -> Tuple[Namespace, List[str]]:
        if args is None:
            # $ prog command <...>
            args = sys.argv[2:]
        else:
            args = list(args)

        return super().parse_known_args(args, namespace)

    def add_help_action(self) -> None:
        self.add_argument("-h", "--help", action="help", help="Show this help message and exit")

    # https://github.com/python/cpython/issues/95073
    def remove_help_action(self) -> None:
        for action in self._actions:
            if not isinstance(action, argparse._HelpAction):
                continue
            self._remove_action(action)
            for option_string in action.option_strings:
                del self._option_string_actions[option_string]



from ._create_vscenario import create_vscenario
from ._multi_scenario_discoverer import MultiScenarioDiscoverer
from ._scenario_discoverer import ScenarioDiscoverer

__all__ = ("ScenarioDiscoverer", "MultiScenarioDiscoverer", "create_vscenario",)



from ._config_file_loader import ConfigFileLoader
from ._config_loader import ConfigLoader
from ._config_type import Config, ConfigType, Section

__all__ = ("Config", "Section", "ConfigType",
           "ConfigLoader", "ConfigFileLoader",)



from ._random import RandomGenerator, StandardRandomGenerator
from ._seeder import Seeder, SeederPlugin

__all__ = ("Seeder", "SeederPlugin",
           "RandomGenerator", "StandardRandomGenerator",)



from abc import abstractmethod
from pathlib import Path
from types import ModuleType

__all__ = ("ModuleLoader",)


class ModuleLoader:
    @abstractmethod
    async def load(self, path: Path) -> ModuleType:
        raise NotImplementedError()



from typing import List

from vedro.core import AggregatedResult, MonotonicScenarioScheduler, ScenarioResult

__all__ = ("RepeaterScenarioScheduler",)


class RepeaterScenarioScheduler(MonotonicScenarioScheduler):
    def aggregate_results(self, scenario_results: List[ScenarioResult]) -> AggregatedResult:
        assert len(scenario_results) > 0

        passed, failed = [], []
        for scenario_result in scenario_results:
            if scenario_result.is_passed():
                passed.append(scenario_result)
            elif scenario_result.is_failed():
                failed.append(scenario_result)

        if len(passed) == 0 and len(failed) == 0:
            scenario_result = scenario_results[-1]
        else:
            scenario_result = failed[-1] if len(failed) > 0 else passed[-1]

        return AggregatedResult.from_existing(scenario_result, scenario_results)



from abc import ABC, abstractmethod
from typing import List

from .._virtual_scenario import VirtualScenario

__all__ = ("ScenarioOrderer",)


class ScenarioOrderer(ABC):
    @abstractmethod
    async def sort(self, scenarios: List[VirtualScenario]) -> List[VirtualScenario]:
        pass



import signal
from types import FrameType
from typing import Any, Dict, Optional, Tuple, Type, Union

from vedro.core import Dispatcher, Plugin, PluginConfig
from vedro.core.scenario_runner import Interrupted
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ScenarioReportedEvent,
    ScenarioRunEvent,
    ScenarioSkippedEvent,
    StartupEvent,
)

__all__ = ("Interrupter", "InterrupterPlugin", "InterrupterPluginTriggered",)


class InterrupterPluginTriggered(Interrupted):
    pass


class InterrupterPlugin(Plugin):
    def __init__(self, config: Type["Interrupter"]) -> None:
        super().__init__(config)
        self._fail_fast: bool = False
        self._failed: int = 0
        self._signals: Tuple[signal.Signals, ...] = config.handle_signals
        self._orig_handlers: Dict[signal.Signals, Any] = {}

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(StartupEvent, self.on_startup) \
                  .listen(ScenarioRunEvent, self.on_scenario_run) \
                  .listen(ScenarioSkippedEvent, self.on_scenario_run) \
                  .listen(ScenarioReportedEvent, self.on_scenario_reported) \
                  .listen(CleanupEvent, self.on_cleanup)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        event.arg_parser.add_argument("--fail-fast", "--ff", "-f",
                                      action="store_true",
                                      default=self._fail_fast,
                                      help="Stop after first failed scenario")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._fail_fast = event.args.fail_fast

    def on_startup(self, event: StartupEvent) -> None:
        if not self._fail_fast:
            return

        def handler(signum: int, frame: Optional[FrameType]) -> None:
            try:
                signame = signal.Signals(signum)
            except ValueError:
                signame = signum  # type: ignore
            raise InterrupterPluginTriggered(f"Stop after signal {signame!r} received")

        for sig in self._signals:
            self._orig_handlers[sig] = signal.getsignal(sig)
            signal.signal(sig, handler)

    def on_scenario_reported(self, event: ScenarioReportedEvent) -> None:
        if event.aggregated_result.is_failed():
            self._failed += 1

    def on_scenario_run(self, event: Union[ScenarioRunEvent, ScenarioSkippedEvent]) -> None:
        if self._fail_fast and self._failed > 0:
            raise InterrupterPluginTriggered("Stop after first failed scenario")

    def on_cleanup(self, event: CleanupEvent) -> None:
        if not self._fail_fast:
            return

        for sig, orig_handler in self._orig_handlers.items():
            signal.signal(sig, orig_handler)
        self._orig_handlers = {}


class Interrupter(PluginConfig):
    plugin = InterrupterPlugin
    description = "Stops test execution after the first failed scenario or on specified signals"

    # Raise Interrupted exception on these signals
    handle_signals: Tuple[signal.Signals, ...] = (signal.SIGTERM,)



from importlib.abc import Loader
from types import ModuleType
from typing import Any

from vedro.core import ScenarioFileLoader

__all__ = ("ScenarioAssertRewriterLoader",)

RewriterType = Any


class ScenarioAssertRewriterLoader(ScenarioFileLoader):
    def __init__(self, rewriter: RewriterType) -> None:
        super().__init__()
        self._rewriter = rewriter

    def _exec_module(self, loader: Loader, module: ModuleType) -> None:
        self._rewriter.exec_module(module)



from typing import List

from vedro.core._virtual_scenario import VirtualScenario

from ._scenario_result import ScenarioResult

__all__ = ("AggregatedResult",)


class AggregatedResult(ScenarioResult):
    def __init__(self, scenario: VirtualScenario) -> None:
        super().__init__(scenario)
        self._scenario_results: List[ScenarioResult] = []

    @property
    def scenario_results(self) -> List[ScenarioResult]:
        return self._scenario_results[:]

    def add_scenario_result(self, scenario_result: ScenarioResult) -> None:
        self._scenario_results.append(scenario_result)

    @staticmethod
    def from_existing(main_scenario_result: ScenarioResult,
                      scenario_results: List[ScenarioResult]) -> "AggregatedResult":
        result = AggregatedResult(main_scenario_result.scenario)

        if main_scenario_result.is_passed():
            result.mark_passed()
        elif main_scenario_result.is_failed():
            result.mark_failed()
        elif main_scenario_result.is_skipped():
            result.mark_skipped()

        if main_scenario_result.started_at is not None:
            result.set_started_at(main_scenario_result.started_at)
        if main_scenario_result.ended_at is not None:
            result.set_ended_at(main_scenario_result.ended_at)

        result.set_scope(main_scenario_result.scope)

        for step_result in main_scenario_result.step_results:
            result.add_step_result(step_result)

        for artifact in main_scenario_result.artifacts:
            result.attach(artifact)

        assert len(scenario_results) > 0
        for scenario_result in scenario_results:
            result.add_scenario_result(scenario_result)

        return result



from ._aggregated_result import AggregatedResult
from ._scenario_result import ScenarioResult
from ._scenario_status import ScenarioStatus

__all__ = ("ScenarioResult", "AggregatedResult", "ScenarioStatus",)



from .orderer_plugin import Orderer, OrdererPlugin
from .random_orderer import RandomOrderer
from .reversed_orderer import ReversedOrderer
from .stable_orderer import StableScenarioOrderer

__all__ = ("Orderer", "OrdererPlugin",
           "StableScenarioOrderer", "ReversedOrderer", "RandomOrderer",)



import ast
import token
from io import BytesIO
from tokenize import tokenize
from typing import Dict, Union

from ._config_markup import (
    ConfigMarkup,
    ConfigSectionType,
    EnabledAttrType,
    ImportType,
    PluginListSectionType,
    PluginSectionType,
)

__all__ = ("ConfigParser",)


class ConfigParser:
    async def parse(self, config_source: str, *, default_indent: str = " " * 4) -> ConfigMarkup:
        config_ast = ast.parse(config_source)
        imports = self._parse_imports(config_ast)
        config_section = self._parse_config_section(config_ast)
        indent = self._get_indent(config_source, default_indent)
        return ConfigMarkup(config_section, imports, indent)

    def _get_indent(self, config_source: str, default_indent: str) -> str:
        stream = BytesIO(config_source.encode())
        indents = [tok.string for tok in tokenize(stream.readline) if tok.type == token.INDENT]
        return min(indents, key=len) if indents else default_indent

    def _parse_imports(self, st: ast.AST) -> Dict[str, ImportType]:
        res: Dict[str, ImportType] = {}
        for node in ast.iter_child_nodes(st):
            if not isinstance(node, ast.Import):
                continue
            end_lineno = getattr(node, "end_lineno", 0)
            for alias in node.names:
                res[alias.name] = {
                    "alias": alias.asname,
                    "start": node.lineno,
                    "end": end_lineno,
                    "offset": node.col_offset,
                }
        return res

    def _parse_config_section(self, st: ast.AST) -> Union[ConfigSectionType, None]:
        for node in ast.iter_child_nodes(st):
            if isinstance(node, ast.ClassDef) and (node.name == "Config"):
                end_lineno = getattr(node, "end_lineno", 0)
                return {
                    "plugins": self._parse_plugin_list_section(node),
                    "start": node.lineno,
                    "end": end_lineno,
                    "offset": node.col_offset,
                }
        return None

    def _parse_plugin_list_section(self, st: ast.ClassDef) -> Union[PluginListSectionType, None]:
        for node in ast.iter_child_nodes(st):
            if isinstance(node, ast.ClassDef) and (node.name == "Plugins"):
                end_lineno = getattr(node, "end_lineno", 0)
                return {
                    "start": node.lineno,
                    "end": end_lineno,
                    "offset": node.col_offset,
                    "children": self._parse_plugin_section(node),
                }
        return None

    def _parse_plugin_section(self, st: ast.ClassDef) -> Dict[str, PluginSectionType]:
        res: Dict[str, PluginSectionType] = {}
        for node in ast.iter_child_nodes(st):
            if isinstance(node, ast.ClassDef):
                end_lineno = getattr(node, "end_lineno", 0)
                res[node.name] = {
                    "start": node.lineno,
                    "end": end_lineno,
                    "offset": node.col_offset,
                    "enabled": self._parse_enabled_attr(node),
                }
        return res

    def _parse_enabled_attr(self, st: ast.ClassDef) -> Union[EnabledAttrType, None]:
        for node in ast.iter_child_nodes(st):
            if not isinstance(node, ast.Assign):
                continue
            target = node.targets[0]
            if isinstance(target, ast.Name) and (target.id == "enabled"):
                end_lineno = getattr(node, "end_lineno", 0)
                return {
                    "start": node.lineno,
                    "end": end_lineno,
                    "offset": node.col_offset,
                }
        return None



from inspect import isclass
from typing import Callable, Optional, Type, TypeVar

from vedro._scenario import Scenario

from ._skip import skip

__all__ = ("skip_if",)


T = TypeVar("T", bound=Type[Scenario])


def skip_if(cond: Callable[[], bool], reason: Optional[str] = None) -> Callable[[T], T]:
    if isclass(cond) or not callable(cond):
        raise TypeError('Usage: @skip_if(<condition>, "reason?")')

    def wrapped(scenario: T) -> T:
        if not issubclass(scenario, Scenario):
            raise TypeError("Decorator @skip_if can be used only with 'vedro.Scenario' subclasses")
        if cond():
            return skip(reason)(scenario)
        return scenario

    return wrapped



from ._plain_scenario_orderer import PlainScenarioOrderer
from ._scenario_orderer import ScenarioOrderer
from ._stable_scenario_orderer import StableScenarioOrderer

__all__ = ("ScenarioOrderer", "PlainScenarioOrderer", "StableScenarioOrderer",)



from vedro.core import ConfigType, Dispatcher, Plugin, PluginConfig
from vedro.core.scenario_orderer import StableScenarioOrderer
from vedro.events import ArgParsedEvent, ArgParseEvent, ConfigLoadedEvent

from .random_orderer import RandomOrderer
from .reversed_orderer import ReversedOrderer

__all__ = ("Orderer", "OrdererPlugin")


class OrdererPlugin(Plugin):
    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ConfigLoadedEvent, self.on_config_loaded) \
            .listen(ArgParseEvent, self.on_arg_parse) \
            .listen(ArgParsedEvent, self.on_arg_parsed)

    def on_config_loaded(self, event: ConfigLoadedEvent) -> None:
        self._global_config: ConfigType = event.config

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        group = event.arg_parser.add_argument_group("Orderer")
        exgroup = group.add_mutually_exclusive_group()

        exgroup.add_argument("--order-stable", action="store_true", default=False,
                             help="Set stable scenario order (default)")
        exgroup.add_argument("--order-reversed", action="store_true", default=False,
                             help="Set reversed scenario order")
        exgroup.add_argument("--order-random", action="store_true", default=False,
                             help="Set random scenario order")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        if event.args.order_random:
            self._global_config.Registry.ScenarioOrderer.register(RandomOrderer, self)
            return

        if event.args.order_reversed:
            self._global_config.Registry.ScenarioOrderer.register(ReversedOrderer, self)
            return

        if event.args.order_stable:
            self._global_config.Registry.ScenarioOrderer.register(StableScenarioOrderer, self)


class Orderer(PluginConfig):
    plugin = OrdererPlugin
    description = "Configures the execution order of scenarios"



import json
import os
from traceback import format_exception
from types import FrameType, TracebackType
from typing import Any, Callable, Dict, List, Optional, Union, cast

from rich.console import Console, RenderableType
from rich.status import Status
from rich.style import Style
from rich.traceback import Trace, Traceback

import vedro
from vedro.core import ExcInfo, ScenarioStatus, StepStatus

__all__ = ("RichPrinter",)


def make_console() -> Console:
    return Console(highlight=False, force_terminal=True, markup=False, soft_wrap=True)


class RichPrinter:
    def __init__(self, console_factory: Callable[[], Console] = make_console,
                 *, traceback_factory: Callable[..., Traceback] = Traceback) -> None:
        self._console = console_factory()
        self._traceback_factory = traceback_factory
        self._scenario_spinner: Union[Status, None] = None

    @property
    def console(self) -> Console:
        return self._console

    def print_header(self, title: str = "Scenarios") -> None:
        self._console.out(title)

    def print_namespace(self, namespace: str) -> None:
        namespace = namespace.replace("_", " ").replace("/", " / ")
        self._console.out(f"* {namespace}", style=Style(bold=True))

    def print_scenario_subject(self, subject: str, status: ScenarioStatus, *,
                               elapsed: Optional[float] = None, prefix: str = "") -> None:
        if status == ScenarioStatus.PASSED:
            subject = f"✔ {subject}"
            style = Style(color="green")
        elif status == ScenarioStatus.FAILED:
            subject = f"✗ {subject}"
            style = Style(color="red")
        elif status == ScenarioStatus.SKIPPED:
            subject = f"○ {subject}"
            style = Style(color="grey70")
        else:
            return

        self._console.out(prefix, end="")
        if elapsed is not None:
            self._console.out(subject, style=style, end="")
            self._console.out(f" ({self.format_elapsed(elapsed)})", style=Style(color="grey50"))
        else:
            self._console.out(subject, style=style)

    def print_scenario_caption(self, caption: str, *, prefix: str = "") -> None:
        self._console.out(prefix, end="")
        self._console.out(f"{caption}", style=Style(color="grey50"))

    def print_step_name(self, name: str, status: StepStatus, *,
                        elapsed: Optional[float] = None, prefix: str = "") -> None:
        if status == StepStatus.PASSED:
            name = f"✔ {name}"
            style = Style(color="green")
        elif status == StepStatus.FAILED:
            name = f"✗ {name}"
            style = Style(color="red")
        else:
            return

        self._console.out(prefix, end="")
        if elapsed is not None:
            self._console.out(name, style=style, end="")
            self._console.out(f" ({self.format_elapsed(elapsed)})", style=Style(color="grey50"))
        else:
            self._console.out(name, style=style)

    def __filter_internals(self, traceback: TracebackType) -> TracebackType:
        class _Traceback:
            def __init__(self, tb_frame: FrameType, tb_lasti: int, tb_lineno: int,
                         tb_next: Optional[TracebackType]) -> None:
                self.tb_frame = tb_frame
                self.tb_lasti = tb_lasti
                self.tb_lineno = tb_lineno
                self.tb_next = tb_next

        tb = _Traceback(traceback.tb_frame, traceback.tb_lasti, traceback.tb_lineno,
                        traceback.tb_next)

        root = os.path.dirname(vedro.__file__)
        while tb.tb_next is not None:
            filename = os.path.abspath(tb.tb_frame.f_code.co_filename)
            if os.path.commonpath([root, filename]) != root:
                break
            tb = tb.tb_next  # type: ignore

        return cast(TracebackType, tb)

    def print_exception(self, exc_info: ExcInfo, *,
                        max_frames: int = 8, show_internal_calls: bool = False) -> None:
        if show_internal_calls:
            traceback = exc_info.traceback
        else:
            traceback = self.__filter_internals(exc_info.traceback)

        formatted = format_exception(exc_info.type, exc_info.value, traceback, limit=max_frames)
        self._console.out("".join(formatted), style=Style(color="yellow"))

    def __filter_locals(self, trace: Trace) -> None:
        for stack in trace.stacks:
            for frame in stack.frames:
                if frame.locals is not None:
                    frame.locals = {k: v for k, v in frame.locals.items()
                                    if k != "self" and k.isidentifier()}

    def print_pretty_exception(self, exc_info: ExcInfo, *,
                               max_frames: int = 8,  # min=4 (see rich.traceback.Traceback impl)
                               show_locals: bool = False,
                               show_internal_calls: bool = False,
                               word_wrap: bool = False) -> None:
        if show_internal_calls:
            traceback = exc_info.traceback
        else:
            traceback = self.__filter_internals(exc_info.traceback)

        trace = Traceback.extract(exc_info.type, exc_info.value, traceback,
                                  show_locals=show_locals)

        if show_locals:
            self.__filter_locals(trace)

        tb = self._traceback_factory(trace, max_frames=max_frames, word_wrap=word_wrap)
        self._console.print(tb)
        self.print_empty_line()

    def pretty_format(self, value: Any) -> Any:
        try:
            return json.dumps(value, ensure_ascii=False, indent=4)
        except BaseException:
            return repr(value)

    def print_scope(self, scope: Dict[str, Any]) -> None:
        self.print_scope_header("Scope")
        for key, val in scope.items():
            self.print_scope_key(key, indent=1)
            self.print_scope_val(self.pretty_format(val))
        self.print_empty_line()

    def print_scope_header(self, title: str) -> None:
        self._console.out(title, style=Style(color="blue", bold=True))

    def print_scope_key(self, key: str, *, indent: int = 0, line_break: bool = False) -> None:
        prepend = " " * indent
        end = "\n" if line_break else ""
        self._console.out(f"{prepend}{key}: ", end=end, style=Style(color="blue"))

    def print_scope_val(self, val: Any) -> None:
        self._console.print(val)

    def print_interrupted(self, exc_info: ExcInfo, *, show_traceback: bool = False) -> None:
        message = f"!!! Interrupted by “{exc_info.value!r}“ !!!"
        spaces = " " * (len(message) - 6)
        multiline_message = "\n".join([
            "!!!" + spaces + "!!!",
            message,
            "!!!" + spaces + "!!!",
        ])
        self._console.out(multiline_message, style=Style(color="yellow"))
        if show_traceback:
            self.print_exception(exc_info)

    def print_report_summary(self, summary: List[str]) -> None:
        if len(summary) == 0:
            return
        text = "# " + "\n# ".join(summary)
        self._console.out(text, style=Style(color="grey70"))

    def format_elapsed(self, elapsed: float) -> str:
        hours = int(elapsed // 3600)
        minutes = int((elapsed - hours * 3600) // 60)
        seconds = elapsed - hours * 3600 - minutes * 60

        formatted = f"{seconds:.2f}s" if elapsed < 60 else f"{int(seconds)}s"
        if (minutes > 0) or (hours > 0):
            formatted = f"{minutes}m {formatted}"
        if hours > 0:
            formatted = f"{hours}h {formatted}"
        return formatted

    def print_report_stats(self, *, total: int, passed: int, failed: int, skipped: int,
                           elapsed: float, is_interrupted: bool = False) -> None:
        if is_interrupted or (failed > 0 or passed == 0):
            style = Style(color="red", bold=True)
        else:
            style = Style(color="green", bold=True)

        scenarios = "scenario" if (total == 1) else "scenarios"
        self._console.out(f"# {total} {scenarios}, "
                          f"{passed} passed, {failed} failed, {skipped} skipped",
                          style=style, end="")
        self._console.out(f" ({self.format_elapsed(elapsed)})", style=Style(color="blue"))

    def print_empty_line(self) -> None:
        self._console.out(" ")

    def print(self, smth: Any, *, end: str = "\n") -> None:
        self._console.out(smth, end=end)

    def show_spinner(self, status: RenderableType = "") -> None:
        if self._scenario_spinner:
            self.hide_spinner()
        self._scenario_spinner = self._console.status(status)
        self._scenario_spinner.start()

    def hide_spinner(self) -> None:
        if self._scenario_spinner:
            self._scenario_spinner.stop()
            self._scenario_spinner = None



from asyncio import CancelledError

import vedro.core as core
import vedro.plugins.artifacted as artifacted
import vedro.plugins.assert_rewriter as assert_rewriter
import vedro.plugins.deferrer as deferrer
import vedro.plugins.director as director
import vedro.plugins.dry_runner as dry_runner
import vedro.plugins.interrupter as interrupter
import vedro.plugins.orderer as orderer
import vedro.plugins.repeater as repeater
import vedro.plugins.rerunner as rerunner
import vedro.plugins.seeder as seeder
import vedro.plugins.skipper as skipper
import vedro.plugins.slicer as slicer
import vedro.plugins.tagger as tagger
import vedro.plugins.terminator as terminator
from vedro.core import (
    Dispatcher,
    Factory,
    MonotonicScenarioRunner,
    MonotonicScenarioScheduler,
    MultiScenarioDiscoverer,
    ScenarioDiscoverer,
    ScenarioFileFinder,
    ScenarioFileLoader,
    ScenarioFinder,
    ScenarioLoader,
    ScenarioOrderer,
    ScenarioRunner,
    ScenarioScheduler,
    Singleton,
)
from vedro.core._scenario_finder._file_filters import (
    AnyFilter,
    DunderFilter,
    ExtFilter,
    HiddenFilter,
)
from vedro.core.scenario_orderer import StableScenarioOrderer

__all__ = ("Config",)


class Config(core.Config):
    class Registry(core.Config.Registry):
        Dispatcher = Singleton[Dispatcher](Dispatcher)

        ScenarioFinder = Factory[ScenarioFinder](lambda: ScenarioFileFinder(
            file_filter=AnyFilter([HiddenFilter(), DunderFilter(), ExtFilter(only=["py"])]),
            dir_filter=AnyFilter([HiddenFilter(), DunderFilter()])
        ))

        ScenarioLoader = Factory[ScenarioLoader](ScenarioFileLoader)

        ScenarioOrderer = Factory[ScenarioOrderer](StableScenarioOrderer)

        ScenarioDiscoverer = Factory[ScenarioDiscoverer](lambda: MultiScenarioDiscoverer(
            finder=Config.Registry.ScenarioFinder(),
            loader=Config.Registry.ScenarioLoader(),
            orderer=Config.Registry.ScenarioOrderer(),
        ))

        ScenarioScheduler = Factory[ScenarioScheduler](MonotonicScenarioScheduler)

        ScenarioRunner = Factory[ScenarioRunner](lambda: MonotonicScenarioRunner(
            dispatcher=Config.Registry.Dispatcher(),
            interrupt_exceptions=(KeyboardInterrupt, SystemExit, CancelledError),
        ))

    class Plugins(core.Config.Plugins):
        class Director(director.Director):
            enabled = True

        class RichReporter(director.RichReporter):
            enabled = True

        class SilentReporter(director.SilentReporter):
            enabled = True

        class PyCharmReporter(director.PyCharmReporter):
            enabled = True

        class Orderer(orderer.Orderer):
            enabled = True

        class Deferrer(deferrer.Deferrer):
            enabled = True

        class Artifacted(artifacted.Artifacted):
            enabled = True

        class Interrupter(interrupter.Interrupter):
            enabled = True

        class Seeder(seeder.Seeder):
            enabled = True

        class Skipper(skipper.Skipper):
            enabled = True

        class Slicer(slicer.Slicer):
            enabled = True

        class Tagger(tagger.Tagger):
            enabled = True

        class Repeater(repeater.Repeater):
            enabled = True

        class Rerunner(rerunner.Rerunner):
            enabled = True

        class AssertRewriter(assert_rewriter.AssertRewriter):
            enabled = True

        class DryRunner(dry_runner.DryRunner):
            enabled = True

        class Terminator(terminator.Terminator):
            enabled = True



from typing import Type, Union

from vedro.core import ConfigType, Dispatcher, Plugin, PluginConfig, ScenarioScheduler
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ConfigLoadedEvent,
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    StartupEvent,
)

from ._scheduler import RepeaterScenarioScheduler

__all__ = ("Repeater", "RepeaterPlugin",)


class RepeaterPlugin(Plugin):
    def __init__(self, config: Type["Repeater"]) -> None:
        super().__init__(config)
        self._repeats: int = 0
        self._global_config: Union[ConfigType, None] = None
        self._scheduler: Union[ScenarioScheduler, None] = None
        self._repeat_scenario_id: Union[str, None] = None

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ConfigLoadedEvent, self.on_config_loaded) \
                  .listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(StartupEvent, self.on_startup) \
                  .listen(ScenarioPassedEvent, self.on_scenario_end) \
                  .listen(ScenarioFailedEvent, self.on_scenario_end) \
                  .listen(CleanupEvent, self.on_cleanup)

    def on_config_loaded(self, event: ConfigLoadedEvent) -> None:
        self._global_config = event.config

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        group = event.arg_parser.add_argument_group("Repeater")
        group.add_argument("-N", "--repeats", type=int, default=1,
                           help="Number of times to repeat scenarios (default: 1)")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._repeats = event.args.repeats
        assert self._repeats >= 1

        if self._repeats <= 1:
            return

        assert self._global_config is not None
        self._global_config.Registry.ScenarioScheduler.register(RepeaterScenarioScheduler, self)

    def on_startup(self, event: StartupEvent) -> None:
        self._scheduler = event.scheduler

    def on_scenario_end(self, event:  Union[ScenarioPassedEvent, ScenarioFailedEvent]) -> None:
        if self._repeats <= 1:
            return

        assert isinstance(self._scheduler, RepeaterScenarioScheduler)

        if self._repeat_scenario_id == event.scenario_result.scenario.unique_id:
            return

        self._repeat_scenario_id = event.scenario_result.scenario.unique_id
        for _ in range(self._repeats - 1):
            self._scheduler.schedule(event.scenario_result.scenario)

    def on_cleanup(self, event: CleanupEvent) -> None:
        if self._repeats > 1:
            event.report.add_summary(f"repeated x{self._repeats}")


class Repeater(PluginConfig):
    plugin = RepeaterPlugin
    description = "Repeat scenarios a specified number of times"



from vedro.core import Artifact, FileArtifact, MemoryArtifact

from ._artifacted import (
    Artifacted,
    ArtifactedPlugin,
    attach_artifact,
    attach_scenario_artifact,
    attach_step_artifact,
)

__all__ = ("Artifacted", "ArtifactedPlugin",
           "attach_artifact", "attach_step_artifact", "attach_scenario_artifact",
           "Artifact", "MemoryArtifact", "FileArtifact",)



import sys
from typing import Dict

if sys.version_info >= (3, 8):
    from importlib.metadata import PackageNotFoundError, metadata
else:
    def metadata(name: str) -> Dict[str, str]:
        raise PackageNotFoundError(name)
    PackageNotFoundError = Exception

from dataclasses import dataclass
from typing import Type

import vedro
from vedro.core import PluginConfig

__all__ = ("get_plugin_info", "PluginInfo",)


@dataclass
class PluginInfo:
    name: str
    enabled: bool
    package: str = "Unknown"
    version: str = "0.0.0"
    summary: str = "No data"
    is_default: bool = False


def get_plugin_info(plugin_config: Type[PluginConfig]) -> PluginInfo:
    plugin_name = getattr(plugin_config, "__name__", "Plugin")
    plugin_info = PluginInfo(plugin_name, plugin_config.enabled)

    plugin = plugin_config.plugin
    module = plugin.__module__
    package = module.split(".")[0]

    # plugin declared in vedro.cfg.py
    if module == "vedro.cfg":
        plugin_info.package = module
        if plugin_config.description:
            plugin_info.summary = plugin_config.description
        return plugin_info

    # default plugin
    if package == "vedro":
        summary = plugin_config.description or "Core Plugin"
        plugin_info.package = ".".join(module.split(".")[:-1])
        plugin_info.version = vedro.__version__
        plugin_info.summary = summary
        plugin_info.is_default = True
        return plugin_info

    try:
        meta = metadata(package)
    except PackageNotFoundError:
        return plugin_info

    plugin_info.package = package
    if "Version" in meta:
        plugin_info.version = meta["Version"]
    if "Summary" in meta:
        plugin_info.summary = meta["Summary"]
    return plugin_info



import random
from abc import ABC, abstractmethod
from typing import Tuple, TypeVar, cast

__all__ = ("StandardRandomGenerator", "RandomGenerator",
           "SeedType", "StateType",)

SeedType = TypeVar("SeedType", int, float, str, bytes, bytearray)
StateType = Tuple[int, Tuple[int], None]


class RandomGenerator(ABC):
    @abstractmethod
    def set_seed(self, seed: SeedType) -> None:
        pass

    @abstractmethod
    def random_int(self, start: int, end: int) -> int:
        pass

    @abstractmethod
    def get_state(self) -> StateType:
        pass

    @abstractmethod
    def set_state(self, state: StateType) -> None:
        pass


class StandardRandomGenerator(RandomGenerator):
    def set_seed(self, seed: SeedType) -> None:
        random.seed(seed)

    def random_int(self, start: int, end: int) -> int:
        return random.randint(start, end)

    def get_state(self) -> StateType:
        return cast(StateType, random.getstate())

    def set_state(self, state: StateType) -> None:
        random.setstate(state)



from abc import ABC, abstractmethod
from typing import Any, Type

from vedro import Config

from ._cmd_arg_parser import CommandArgumentParser

__all__ = ("Command",)


class Command(ABC):
    def __init__(self, config: Type[Config],
                 arg_parser: CommandArgumentParser, **kwargs: Any) -> None:
        self._config = config
        self._arg_parser = arg_parser

    @abstractmethod
    async def run(self) -> None:
        pass



from typing import Tuple, Type

from vedro.core import ConfigType, Dispatcher, Plugin, PluginConfig
from vedro.events import ArgParsedEvent, ArgParseEvent, ConfigLoadedEvent

from ._dry_runner import DryRunner as DryRunnerImpl


class DryRunnerPlugin(Plugin):
    def __init__(self, config: Type["DryRunner"]) -> None:
        super().__init__(config)
        self._dry_run = False
        self._interrupt_exceptions = config.interrupt_exceptions

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ConfigLoadedEvent, self.on_config_loaded) \
                  .listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed)

    def on_config_loaded(self, event: ConfigLoadedEvent) -> None:
        self._global_config: ConfigType = event.config

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        group = event.arg_parser.add_argument_group("Dry Runner")
        group.add_argument("--dry-run", action="store_true", default=self._dry_run,
                           help="Run scenarios without executing them")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._dry_run = event.args.dry_run
        if self._dry_run:
            self._global_config.Registry.ScenarioRunner.register(
                resolver=lambda: DryRunnerImpl(self._global_config.Registry.Dispatcher(),
                                               interrupt_exceptions=self._interrupt_exceptions),
                registrant=self
            )


class DryRunner(PluginConfig):
    plugin = DryRunnerPlugin
    description = "Simulates scenario execution without actually executing them"

    # Exceptions that will interrupt scenario execution
    interrupt_exceptions: Tuple[Type[BaseException], ...] = (KeyboardInterrupt, SystemExit,)



from asyncio import iscoroutinefunction
from typing import Any, Callable

__all__ = ("VirtualStep",)


class VirtualStep:
    def __init__(self, orig_step: Callable[..., None]) -> None:
        self._orig_step = orig_step

    @property
    def name(self) -> str:
        return self._orig_step.__name__

    def is_coro(self) -> bool:
        return iscoroutinefunction(self._orig_step)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._orig_step(*args, **kwargs)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self._orig_step.__name__!r}>"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and (self.__dict__ == other.__dict__)



from inspect import isclass
from typing import Callable, Type, TypeVar, overload

from vedro._scenario import Scenario

__all__ = ("only",)

T = TypeVar("T", bound=Type[Scenario])


@overload
def only(scenario_or_reason: T) -> T:  # pragma: no cover
    pass


@overload
def only(scenario_or_reason: None = None) -> Callable[[T], T]:  # pragma: no cover
    pass


def only(scenario_or_nothing=None):  # type: ignore
    def wrapped(scenario: T) -> T:
        if not issubclass(scenario, Scenario):
            raise TypeError("Decorator @only can be used only with 'vedro.Scenario' subclasses")
        setattr(scenario, "__vedro__only__", True)
        return scenario

    if scenario_or_nothing is None:
        return wrapped
    elif isclass(scenario_or_nothing):
        return wrapped(scenario_or_nothing)
    else:
        raise TypeError("Usage: @only")



from typing import List

__all__ = ("ConfigGenerator",)


class ConfigGenerator:
    def __init__(self, indent: str) -> None:
        self._indent = indent

    def gen_config_section(self, plugins_section: List[str], *, offset: int = 0) -> List[str]:
        return [
            f"{self._indent * offset}class Config(vedro.Config):",
            *plugins_section,
        ]

    def gen_plugins_section(self, plugin_section: List[str], *, offset: int = 1) -> List[str]:
        return [
            "",
            f"{self._indent * offset}class Plugins(vedro.Config.Plugins):",
            *plugin_section,
        ]

    def gen_plugin_section(self, plugin_package: str, plugin_name: str, enabled: bool, *,
                           offset: int = 2) -> List[str]:
        return [
            "",
            f"{self._indent * offset}class {plugin_name}({plugin_package}.{plugin_name}):",
            f"{self._indent * (offset + 1)}enabled = {enabled}",
        ]

    def gen_enabled_attr(self, enabled: bool, *,
                         offset: int = 3) -> List[str]:
        return [
            f"{self._indent * offset}enabled = {enabled}"
        ]

    def gen_import(self, package: str) -> List[str]:
        return [
            f"import {package}",
        ]



from pathlib import Path
from typing import Type

from vedro import Config
from vedro.core import Dispatcher, Plugin
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ConfigLoadedEvent,
    StartupEvent,
)

from .._cmd_arg_parser import CommandArgumentParser
from .._command import Command

__all__ = ("RunCommand",)


class RunCommand(Command):
    def __init__(self, config: Type[Config], arg_parser: CommandArgumentParser) -> None:
        super().__init__(config, arg_parser)

    async def _register_plugins(self, dispatcher: Dispatcher) -> None:
        for _, section in self._config.Plugins.items():
            assert issubclass(section.plugin, Plugin)
            assert section.plugin != Plugin
            if section.enabled:
                plugin = section.plugin(config=section)
                dispatcher.register(plugin)

    async def _parse_args(self, dispatcher: Dispatcher) -> None:
        # https://github.com/python/cpython/issues/95073
        self._arg_parser.remove_help_action()
        await dispatcher.fire(ArgParseEvent(self._arg_parser))
        self._arg_parser.add_help_action()

        args = self._arg_parser.parse_args()
        await dispatcher.fire(ArgParsedEvent(args))

    async def run(self) -> None:
        dispatcher = self._config.Registry.Dispatcher()
        await self._register_plugins(dispatcher)

        await dispatcher.fire(ConfigLoadedEvent(self._config.path, self._config))

        await self._parse_args(dispatcher)

        start_dir = Path("scenarios")
        discoverer = self._config.Registry.ScenarioDiscoverer()
        try:
            scenarios = await discoverer.discover(start_dir)
        except SystemExit as e:
            raise Exception(f"SystemExit({e.code}) ⬆")

        scheduler = self._config.Registry.ScenarioScheduler(scenarios)
        await dispatcher.fire(StartupEvent(scheduler))

        runner = self._config.Registry.ScenarioRunner()
        report = await runner.run(scheduler)

        await dispatcher.fire(CleanupEvent(report))



from typing import Any, Type, Union

from vedro.core import Dispatcher, Plugin, PluginConfig, VirtualScenario
from vedro.events import ArgParsedEvent, ArgParseEvent, StartupEvent

__all__ = ("Tagger", "TaggerPlugin",)


class TaggerPlugin(Plugin):
    def __init__(self, config: Type["Tagger"]) -> None:
        super().__init__(config)
        self._tags: Union[str, None] = None

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(StartupEvent, self.on_startup)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        event.arg_parser.add_argument("-t", "--tags", nargs="?", help="Set tag")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._tags = event.args.tags

    def _validate_tags(self, scenario: VirtualScenario, tags: Any) -> bool:
        if not isinstance(tags, (list, tuple, set)):
            raise TypeError(f"Scenario '{scenario.rel_path}' tags must be a list, tuple or set, "
                            f"got {type(tags)}")

        for tag in tags:
            if not tag.isidentifier():
                raise ValueError(
                    f"Scenario '{scenario.rel_path}' tag '{tag}' is not a valid identifier")

        return True

    async def on_startup(self, event: StartupEvent) -> None:
        if self._tags is None:
            return
        async for scenario in event.scheduler:
            tags = getattr(scenario._orig_scenario, "tags", ())
            assert self._validate_tags(scenario, tags)

            if self._tags not in tags:
                event.scheduler.ignore(scenario)


class Tagger(PluginConfig):
    plugin = TaggerPlugin
    description = "Allows scenarios to be selectively run based on user-defined tags"



from abc import ABC, abstractmethod

from .._report import Report
from ..scenario_scheduler import ScenarioScheduler

__all__ = ("ScenarioRunner",)


class ScenarioRunner(ABC):
    @abstractmethod
    async def run(self, scheduler: ScenarioScheduler) -> Report:
        pass



from ._artifacts import Artifact, FileArtifact, MemoryArtifact
from ._config_loader import Config, ConfigFileLoader, ConfigLoader, ConfigType, Section
from ._container import Container, Factory, FactoryType, Singleton
from ._dispatcher import Dispatcher, Subscriber
from ._event import Event
from ._exc_info import ExcInfo
from ._module_loader import ModuleFileLoader, ModuleLoader
from ._plugin import Plugin, PluginConfig
from ._report import Report
from ._scenario_discoverer import MultiScenarioDiscoverer, ScenarioDiscoverer
from ._scenario_finder import ScenarioFileFinder, ScenarioFinder
from ._scenario_loader import ScenarioFileLoader, ScenarioLoader
from ._scenario_result import AggregatedResult, ScenarioResult, ScenarioStatus
from ._step_result import StepResult, StepStatus
from ._virtual_scenario import VirtualScenario
from ._virtual_step import VirtualStep
from .scenario_orderer import ScenarioOrderer
from .scenario_runner import MonotonicScenarioRunner, ScenarioRunner
from .scenario_scheduler import MonotonicScenarioScheduler, ScenarioScheduler

__all__ = ("Dispatcher", "Subscriber", "Event", "ExcInfo", "Plugin", "PluginConfig",
           "Report", "ScenarioRunner", "MonotonicScenarioRunner", "ScenarioDiscoverer",
           "MultiScenarioDiscoverer", "ScenarioFinder", "ScenarioFileFinder", "ScenarioLoader",
           "ScenarioFileLoader", "ScenarioResult", "AggregatedResult", "StepResult",
           "VirtualScenario", "VirtualStep", "ScenarioStatus", "StepStatus",
           "ConfigLoader", "ConfigFileLoader", "Config", "Section", "ConfigType",
           "ModuleLoader", "ModuleFileLoader", "Artifact", "MemoryArtifact", "FileArtifact",
           "ScenarioScheduler", "MonotonicScenarioScheduler", "FactoryType",
           "Container", "Factory", "Singleton", "ScenarioOrderer")



from typing import Any, List, Union, cast

from ._exc_info import ExcInfo
from ._scenario_result import AggregatedResult

__all__ = ("Report",)


class Report:
    def __init__(self) -> None:
        self._summary: List[str] = []
        self._started_at: Union[float, None] = None
        self._ended_at: Union[float, None] = None
        self._total: int = 0
        self._passed: int = 0
        self._failed: int = 0
        self._skipped: int = 0
        self._interrupted: Union[ExcInfo, None] = None

    @property
    def interrupted(self) -> Union[ExcInfo, None]:
        return self._interrupted

    @property
    def started_at(self) -> Union[float, None]:
        return self._started_at

    @property
    def ended_at(self) -> Union[float, None]:
        return self._ended_at

    @property
    def total(self) -> int:
        return self._total

    @property
    def passed(self) -> int:
        return self._passed

    @property
    def failed(self) -> int:
        return self._failed

    @property
    def skipped(self) -> int:
        return self._skipped

    @property
    def summary(self) -> List[str]:
        return self._summary[:]

    @property
    def elapsed(self) -> float:
        if (self.ended_at is None) or (self.started_at is None):
            return 0.0
        return self.ended_at - self.started_at

    def add_result(self, result: AggregatedResult) -> None:
        self._total += 1
        if result.is_passed():
            self._passed += 1
        elif result.is_failed():
            self._failed += 1
        elif result.is_skipped():
            self._skipped += 1

        if result.started_at:
            if self.started_at is None:
                self._started_at = result.started_at
            self._started_at = min(cast(float, self._started_at), result.started_at)

        if result.ended_at:
            if self.ended_at is None:
                self._ended_at = result.ended_at
            self._ended_at = max(cast(float, self._ended_at), result.ended_at)

    def add_summary(self, summary: str) -> None:
        self._summary.append(summary)

    def set_interrupted(self, exc_info: ExcInfo) -> None:
        self._interrupted = exc_info

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and (self.__dict__ == other.__dict__)

    def __repr__(self) -> str:
        interrupted = self.interrupted.type if self.interrupted else None
        return (f"<{self.__class__.__name__} "
                f"total={self._total} passed={self._passed} "
                f"failed={self._failed} skipped={self._skipped} "
                f"interrupted={interrupted}>")



from typing import Dict, Type, Union

from vedro.core import Dispatcher, Plugin, PluginConfig
from vedro.events import ArgParseEvent, ConfigLoadedEvent

from ._director_init_event import DirectorInitEvent
from ._reporter import Reporter

__all__ = ("Director", "DirectorPlugin",)


class DirectorPlugin(Plugin):
    def __init__(self, config: Type["Director"]) -> None:
        super().__init__(config)
        self._dispatcher: Union[Dispatcher, None] = None
        self._reporters: Dict[str, Reporter] = {}
        self._default_reporter = ""

    def subscribe(self, dispatcher: Dispatcher) -> None:
        self._dispatcher = dispatcher.listen(ConfigLoadedEvent, self.on_config_loaded) \
                                     .listen(ArgParseEvent, self.on_arg_parse)

    async def on_config_loaded(self, event: ConfigLoadedEvent) -> None:
        assert isinstance(self._dispatcher, Dispatcher)
        await self._dispatcher.fire(DirectorInitEvent(self))

    async def on_arg_parse(self, event: ArgParseEvent) -> None:
        if len(self._reporters) == 0:
            return
        event.arg_parser.add_argument("-r", "--reporters", nargs='*', choices=self._reporters,
                                      default=[self._default_reporter],
                                      help=f"Set reporter (default {self._default_reporter})")
        args, *_ = event.arg_parser.parse_known_args()
        for reporter_name in args.reporters:
            reporter = self._reporters[reporter_name]
            reporter.on_chosen()

    def register(self, name: str, reporter: Reporter) -> None:
        if len(self._reporters) == 0:
            self._default_reporter = name
        self._reporters[name] = reporter


class Director(PluginConfig):
    plugin = DirectorPlugin
    description = "Manages and configures reporters for scenario execution"



from abc import ABC, abstractmethod
from pathlib import Path
from typing import List

from .._scenario_finder import ScenarioFinder
from .._scenario_loader import ScenarioLoader
from .._virtual_scenario import VirtualScenario
from ..scenario_orderer import ScenarioOrderer

__all__ = ("ScenarioDiscoverer",)


class ScenarioDiscoverer(ABC):
    def __init__(self, finder: ScenarioFinder, loader: ScenarioLoader,
                 orderer: ScenarioOrderer) -> None:
        self._finder = finder
        self._loader = loader
        self._orderer = orderer

    @abstractmethod
    async def discover(self, root: Path) -> List[VirtualScenario]:
        pass



from typing import Callable, Type, Union

from vedro.core import Dispatcher, ExcInfo, PluginConfig, ScenarioResult
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ScenarioReportedEvent,
    ScenarioRunEvent,
    ScenarioSkippedEvent,
    StartupEvent,
)
from vedro.plugins.director._director_init_event import DirectorInitEvent
from vedro.plugins.director._reporter import Reporter

from ._rich_printer import RichPrinter

__all__ = ("RichReporterPlugin", "RichReporter",)


class RichReporterPlugin(Reporter):
    def __init__(self, config: Type["RichReporter"], *,
                 printer_factory: Callable[[], RichPrinter] = RichPrinter) -> None:
        super().__init__(config)
        self._printer = printer_factory()
        self._verbosity = 0
        self._tb_pretty = config.tb_pretty
        self._tb_show_internal_calls = config.tb_show_internal_calls
        self._tb_show_locals = config.tb_show_locals
        self._tb_max_frames = config.tb_max_frames
        self._show_skipped = config.show_skipped
        self._show_timings = config.show_timings
        self._show_paths = config.show_paths
        self._show_steps = config.show_steps
        self._hide_namespaces = config.hide_namespaces
        self._show_scenario_spinner = config.show_scenario_spinner
        self._show_interrupted_traceback = config.show_interrupted_traceback
        self._v2_verbosity = config.v2_verbosity
        self._namespace: Union[str, None] = None

    def subscribe(self, dispatcher: Dispatcher) -> None:
        super().subscribe(dispatcher)
        dispatcher.listen(DirectorInitEvent, lambda e: e.director.register("rich", self))

    def on_chosen(self) -> None:
        assert isinstance(self._dispatcher, Dispatcher)

        self._dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                        .listen(ArgParsedEvent, self.on_arg_parsed) \
                        .listen(StartupEvent, self.on_startup) \
                        .listen(ScenarioRunEvent, self.on_scenario_run) \
                        .listen(ScenarioSkippedEvent, self.on_scenario_skipped) \
                        .listen(ScenarioReportedEvent, self.on_scenario_reported) \
                        .listen(CleanupEvent, self.on_cleanup)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        group = event.arg_parser.add_argument_group("Rich Reporter")

        help_message = ("Increase verbosity level "
                        "(-v: show steps, -vv: show exception, -vvv: show scope)")
        group.add_argument("-v", "--verbose",
                           action="count",
                           default=self._verbosity,
                           help=help_message)
        group.add_argument("--show-timings",
                           action="store_true",
                           default=self._show_timings,
                           help="Show the elapsed time of each scenario")
        group.add_argument("--show-steps",
                           action="store_true",
                           default=self._show_steps,
                           help="Show scenario step names")
        group.add_argument("--show-paths",
                           action="store_true",
                           default=self._show_paths,
                           help="Show the relative path of each passed scenario")
        group.add_argument("--hide-namespaces",
                           action="store_true",
                           default=self._hide_namespaces,
                           help="Don't show scenario namespaces")
        group.add_argument("--tb-show-internal-calls",
                           action="store_true",
                           default=self._tb_show_internal_calls,
                           help="Show internal calls in the traceback output")
        group.add_argument("--tb-show-locals",
                           action="store_true",
                           default=self._tb_show_locals,
                           help="Show local variables in the traceback output")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._verbosity = event.args.verbose
        if self._v2_verbosity:
            self._verbosity = self._verbosity + 2
        self._show_timings = event.args.show_timings
        self._show_paths = event.args.show_paths
        self._show_steps = event.args.show_steps
        self._hide_namespaces = event.args.hide_namespaces
        self._tb_show_internal_calls = event.args.tb_show_internal_calls
        self._tb_show_locals = event.args.tb_show_locals

        assert self._tb_max_frames >= 4, \
            "RichReporter: `max_frames` must be >= 4"

        if self._tb_show_locals:
            assert self._tb_pretty, \
                "RichReporter: to enable `tb_show_locals` set `tb_pretty` to `True`"

    def on_startup(self, event: StartupEvent) -> None:
        self._printer.print_header()

    def on_scenario_run(self, event: ScenarioRunEvent) -> None:
        namespace = event.scenario_result.scenario.namespace
        if namespace != self._namespace:
            self._namespace = namespace
            if self._hide_namespaces is False:
                self._printer.print_namespace(namespace)

        if self._show_scenario_spinner:
            self._printer.show_spinner(f" {event.scenario_result.scenario.subject}")

    def on_scenario_skipped(self, event: ScenarioSkippedEvent) -> None:
        if not self._show_skipped:
            return
        namespace = event.scenario_result.scenario.namespace
        if namespace != self._namespace:
            self._namespace = namespace
            if self._hide_namespaces is False:
                self._printer.print_namespace(namespace)

    def _print_exception(self, exc_info: ExcInfo) -> None:
        if self._tb_pretty:
            self._printer.print_pretty_exception(exc_info,
                                                 max_frames=self._tb_max_frames,
                                                 show_locals=self._tb_show_locals,
                                                 show_internal_calls=self._tb_show_internal_calls)
        else:
            self._printer.print_exception(exc_info,
                                          max_frames=self._tb_max_frames,
                                          show_internal_calls=self._tb_show_internal_calls)

    def _prefix_to_indent(self, prefix: str, indent: int = 0) -> str:
        last_line = prefix.split("\n")[-1]
        return (len(last_line) + indent) * " "

    def _print_scenario_passed(self, scenario_result: ScenarioResult, *, prefix: str = "") -> None:
        elapsed = scenario_result.elapsed if self._show_timings else None
        self._printer.print_scenario_subject(scenario_result.scenario.subject,
                                             scenario_result.status,
                                             elapsed=elapsed,
                                             prefix=prefix)
        if self._show_steps:
            for step_result in scenario_result.step_results:
                elapsed = step_result.elapsed if self._show_timings else None
                step_prefix = self._prefix_to_indent(prefix, indent=2)
                self._printer.print_step_name(step_result.step_name, step_result.status,
                                              elapsed=elapsed, prefix=step_prefix)

        if self._show_paths:
            caption = f"> {scenario_result.scenario.rel_path}"
            caption_prefix = self._prefix_to_indent(prefix, indent=2)
            self._printer.print_scenario_caption(caption, prefix=caption_prefix)

    def _print_scenario_failed(self, scenario_result: ScenarioResult, *, prefix: str = "") -> None:
        elapsed = scenario_result.elapsed if self._show_timings else None
        self._printer.print_scenario_subject(scenario_result.scenario.subject,
                                             scenario_result.status,
                                             elapsed=elapsed,
                                             prefix=prefix)

        if self._verbosity > 0:
            for step_result in scenario_result.step_results:
                elapsed = step_result.elapsed if self._show_timings else None
                step_prefix = self._prefix_to_indent(prefix, indent=2)
                self._printer.print_step_name(step_result.step_name, step_result.status,
                                              elapsed=elapsed, prefix=step_prefix)

        if self._verbosity > 1:
            for step_result in scenario_result.step_results:
                if step_result.exc_info:
                    self._print_exception(step_result.exc_info)

        if self._verbosity > 2:
            if scenario_result.scope:
                self._printer.print_scope(scenario_result.scope)

    def _print_scenario_skipped(self, scenario_result: ScenarioResult, *,
                                prefix: str = "") -> None:
        if not self._show_skipped:
            return
        elapsed = scenario_result.elapsed if self._show_timings else None
        self._printer.print_scenario_subject(scenario_result.scenario.subject,
                                             scenario_result.status,
                                             elapsed=elapsed,
                                             prefix=prefix)

    def _print_scenario_result(self, scenario_result: ScenarioResult, *, prefix: str = "") -> None:
        if scenario_result.is_passed():
            self._print_scenario_passed(scenario_result, prefix=prefix)
        elif scenario_result.is_failed():
            self._print_scenario_failed(scenario_result, prefix=prefix)
        elif scenario_result.is_skipped():
            self._print_scenario_skipped(scenario_result, prefix=prefix)

    def on_scenario_reported(self, event: ScenarioReportedEvent) -> None:
        if self._show_scenario_spinner:
            self._printer.hide_spinner()

        aggregated_result = event.aggregated_result
        rescheduled = len(aggregated_result.scenario_results)
        if rescheduled == 1:
            self._print_scenario_result(aggregated_result, prefix=" ")
            return

        self._printer.print_scenario_subject(aggregated_result.scenario.subject,
                                             aggregated_result.status, elapsed=None, prefix=" ")
        for index, scenario_result in enumerate(aggregated_result.scenario_results, start=1):
            prefix = f" │\n ├─[{index}/{rescheduled}] "
            self._print_scenario_result(scenario_result, prefix=prefix)

        self._printer.print_empty_line()

    def on_cleanup(self, event: CleanupEvent) -> None:
        self._printer.print_empty_line()

        is_interrupted = False
        if event.report.interrupted:
            is_interrupted = True
            self._printer.print_interrupted(event.report.interrupted,
                                            show_traceback=self._show_interrupted_traceback)

        self._printer.print_report_summary(event.report.summary)
        self._printer.print_report_stats(total=event.report.total,
                                         passed=event.report.passed,
                                         failed=event.report.failed,
                                         skipped=event.report.skipped,
                                         elapsed=event.report.elapsed,
                                         is_interrupted=is_interrupted)


class RichReporter(PluginConfig):
    plugin = RichReporterPlugin
    description = "Enhanced, customizable scenario reporting with rich output"

    # Show skipped scenarios
    show_skipped: bool = True

    # Show the elapsed time of each scenario
    show_timings: bool = False

    # Show the relative path of each passed scenario
    show_paths: bool = False

    # Show scenario step names
    show_steps: bool = False

    # Don't show scenario namespaces
    hide_namespaces: bool = False

    # Show status indicator of the current running scenario
    show_scenario_spinner: bool = False

    # Show pretty traceback
    tb_pretty: bool = True

    # Show internal calls in the traceback output
    tb_show_internal_calls: bool = False

    # Show local variables in the traceback output
    # Available if tb_pretty is True
    tb_show_locals: bool = False

    # Max stack trace entries to show (min=4)
    tb_max_frames: int = 8

    # Show traceback if the execution is interrupted
    show_interrupted_traceback: bool = False

    # Enable new verbose levels
    v2_verbosity: bool = False



import uuid
from typing import Dict, Type, Union

from vedro.core import Dispatcher, Plugin, PluginConfig
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ScenarioRunEvent,
    StartupEvent,
)

from ._random import RandomGenerator, StandardRandomGenerator, StateType

__all__ = ("Seeder", "SeederPlugin",)

_random = StandardRandomGenerator()


class SeederPlugin(Plugin):
    MIN_SEED = 1
    MAX_SEED = 2 ** 63 - 1

    def __init__(self, config: Type["Seeder"], *, random: RandomGenerator = _random) -> None:
        super().__init__(config)
        self._random = random

        self._inital_seed: Union[str, None] = None
        self._discovered_seed: Union[int, None] = None
        self._scheduled_seed: Union[int, None] = None
        self._scheduled_state: Union[StateType, None] = None

        self._scenarios: Dict[str, int] = {}
        self._history: Dict[str, int] = {}

    def _generate_seed(self) -> int:
        return self._random.random_int(self.MIN_SEED, self.MAX_SEED)

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(StartupEvent, self.on_startup) \
                  .listen(ScenarioRunEvent, self.on_scenario_run) \
                  .listen(CleanupEvent, self.on_cleanup)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        event.arg_parser.add_argument("--seed", nargs="?", help="Set seed")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._inital_seed = event.args.seed if (event.args.seed is not None) else str(uuid.uuid4())

    def on_startup(self, event: StartupEvent) -> None:
        assert self._inital_seed is not None
        self._random.set_seed(self._inital_seed)

        self._scheduled_seed = self._generate_seed()
        self._random.set_seed(self._scheduled_seed)
        self._scheduled_state = self._random.get_state()

        self._discovered_seed = self._generate_seed()
        self._random.set_seed(self._discovered_seed)

        for scenario in event.scheduler.discovered:
            self._scenarios[scenario.unique_id] = self._generate_seed()

    def on_scenario_run(self, event: ScenarioRunEvent) -> None:
        unique_id = event.scenario_result.scenario.unique_id
        if unique_id not in self._history:
            self._history[unique_id] = 0
        self._history[unique_id] += 1

        if unique_id not in self._scenarios:
            assert self._scheduled_state is not None
            self._random.set_state(self._scheduled_state)
            self._scenarios[unique_id] = self._generate_seed()
            self._scheduled_state = self._random.get_state()

        seed = self._scenarios[unique_id]
        self._random.set_seed(seed)

        for _ in range(self._history[unique_id]):
            seed = self._generate_seed()
        self._random.set_seed(seed)

    def on_cleanup(self, event: CleanupEvent) -> None:
        if (event.report.passed + event.report.failed) > 0:
            event.report.add_summary(f"--seed {self._inital_seed}")


class Seeder(PluginConfig):
    plugin = SeederPlugin
    description = "Sets seeds for deterministic random behavior in scenarios"
